/**
  *  \file client/dialogs/commandedit.cc
  *  \brief Command Edit Dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogbinding.h"
#include "client/dialogs/commandedit.h"
#include "client/screens.h"
#include "client/widgets/commandinfo.h"
#include "client/widgets/commandlist.h"
#include "client/widgets/helpwidget.h"
#include "game/global.h"
#include "game/mine.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets/quit.h"

namespace {
    class WCommandEditDialog : public UIWindow, public IntUserInterfaceDialogBinding {
        ObjectHolder h;
        WCommandList list;
        WCommandInfo info;
        UIButton goto_button;

        void buildDialog();
        void onScroll();
        void onGoto();
     public:
        WCommandEditDialog();
        ~WCommandEditDialog();

        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

WCommandEditDialog::WCommandEditDialog()
    : UIWindow(_("Auxiliary Commands"), BLUE_WINDOW, COLOR_GRAY),
      list(getDisplayedTurn().getCommands(getPlayerId())),
      info(),
      goto_button(1, _("Go to"), 'g')
{
    buildDialog();
    onScroll();
    list.sig_moved.add(this, &WCommandEditDialog::onScroll);
}

WCommandEditDialog::~WCommandEditDialog()
{ }

void
WCommandEditDialog::buildDialog()
{
    // VBox
    //   Frame
    //     WCommandList
    //     UIScrollbar
    //   HBox
    //     WCommandInfo
    //     VBox
    //       UIButton "G"
    //       UISpacer
    //   HBox
    //     UIButton "Close"
    //     UIButton "Del"
    //     UIButton "Ins"
    //     UISpacer
    //     UIButton "Help"

    setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& group0 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    group0.add(list);
    group0.add(h.add(new UIScrollbar(list)));

    UIGroup& group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group11 = h.add(new UIGroup(UIVBoxLayout::instance0));
    group1.add(info);
    group1.add(group11);
    group11.add(goto_button);
    group11.add(h.add(new UISpacer()));

    UIGroup& group2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group2.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    group2.add(h.add(new UIButton(_("Del"),   SDLK_DELETE)));
    group2.add(h.add(new UIButton(_("Ins"),   SDLK_INSERT)));
    group2.add(h.add(new UISpacer()));
    group2.add(h.add(new UIButton(_("Help"),  'h')));

    add(group0);
    add(group1);
    add(group2);
    add(h.add(new WHelpWidget("pcc2:auxcmds")));
    add(h.add(new UIQuit(1)));

    goto_button.sig_fired.add(this, &WCommandEditDialog::onGoto);
    list.setFocusedWidget();
    pack();
}

void
WCommandEditDialog::onScroll()
{
    int id;
    const GCommand* cmd = list.getCurrentCommand();
    info.setCommand(cmd);
    goto_button.setState(st_Disabled,
                         !(cmd != 0
                           && (((id = cmd->getAffectedShip()) != 0 && current_anyship_type.isValidIndex(id))
                               || ((id = cmd->getAffectedPlanet()) != 0 && current_anyplanet_type.isValidIndex(id))
                               || ((id = cmd->getAffectedMinefield()) != 0 && current_minefield_type.isValidIndex(id)))));
}

void
WCommandEditDialog::onGoto()
{
    IntUserInterfaceBinding& ub = IntUserInterfaceBinding::get();
    const GCommand* cmd = list.getCurrentCommand();
    int id;
    if (cmd == 0) {
        /* Ignore */
    } else if ((id = cmd->getAffectedShip()) != 0 && current_anyship_type.isValidIndex(id)) {
        if (current_ship_type.isValidIndex(id)) {
            postGoToScreen(ub, 1, id);
        } else {
            GMapObject& mo = dynamic_cast<GMapObject&>(current_anyship_type.getObjectByIndex(id));
            postGoToChart(ub, mo.getPos(), &mo);
        }
    } else if ((id = cmd->getAffectedPlanet()) != 0 && current_anyplanet_type.isValidIndex(id)) {
        if (current_planet_type.isValidIndex(id)) {
            postGoToScreen(ub, 2, id);
        } else {
            GMapObject& mo = dynamic_cast<GMapObject&>(current_anyplanet_type.getObjectByIndex(id));
            postGoToChart(ub, mo.getPos(), &mo);
        }
    } else if ((id = cmd->getAffectedMinefield()) != 0 && current_minefield_type.isValidIndex(id)) {
        GMapObject& mo = dynamic_cast<GMapObject&>(current_minefield_type.getObjectByIndex(id));
        postGoToChart(ub, mo.getPos(), &mo);
    } else {
        /* what else? */
    }
}

bool
WCommandEditDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == SDLK_RETURN) {
        onGoto();
        return true;
    } else {
        return UIWindow::handleEvent(event, second_pass);
    }
}

/** Edit commands. Displays a dialog to edit commands from the currently displayed turn. */
void
editCommands()
{
    WCommandEditDialog edit;
    ui_root->centerWidget(edit);
    edit.runModal(edit);
}
