/**
  *  \file client/dialogs/fleetcost.cc
  *  \brief Simulator Fleet Cost Comparison
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/fleetcost.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "client/scr-sim.h"
#include "game/config.h"
#include "gfx/palette.h"
#include "game/team.h"
#include "ui/widgets/button.h"
#include "game/formula.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "game/racename.h"
#include "client/dialogs/fleetopts.h"
#include "game/planet.h"
#include "client/widgets/helpwidget.h"
#include "game/planetform.h"
#include "client/widgets/simlist.h"
#include "ui/widgets/quit.h"

static WFleetCostOptions cost_opts = {
    WFleetCostOptions::ff_LFM,
    true,
    false,
    true,
    true,
    false,
    WFleetCostOptions::st_No,
    false,
};

static int
getNextPlayer(GPlayerSet set, int player, int delta)
{
    for (int i = 0; i < NUM_OWNERS; ++i) {
        player += delta;
        if (player > NUM_OWNERS)
            player = 1;
        if (player <= 0)
            player = NUM_OWNERS;
        if (set.contains(player))
            break;
    }
    return player;
}

static void
addTech(GCost& cost, int player, int& haveTech, int needTech)
{
    switch (cost_opts.shiptech) {
     case WFleetCostOptions::st_No:
     case WFleetCostOptions::st_MAX:
        break;

     case WFleetCostOptions::st_Player:
        if (needTech > haveTech) {
            cost.add(el_Money, getBaseTechCost(player, haveTech, needTech));
            haveTech = needTech;
        }
        break;

     case WFleetCostOptions::st_Ship:
        cost.add(el_Money, getBaseTechCost(player, 1, needTech));
        break;
    }
}

static void
addFighters(GCost& cost, int owner, int count)
{
    if (cost_opts.fighters == cost_opts.ff_LFM) {
        cost += config.ShipFighterCost(owner) * count;
    } else if (cost_opts.fighters == cost_opts.ff_Build) {
        cost += config.BaseFighterCost(owner) * count;
    }
}



WFleetCostDialog::WFleetCostDialog(const GSimState& state)
    : UIWindow(_("Fleet Cost Comparison"), BLUE_WINDOW, COLOR_GRAY),
      state(state)
{
    init();
}

WFleetCostDialog::~WFleetCostDialog()
{ }

void
WFleetCostDialog::init()
{
    /* Figure out players taking part */
    for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i)
        available_players += state.getShip(i).getOwner();
    if (state.hasPlanet())
        available_players += state.getPlanet().getOwner();

    /* Figure out teams taking part */
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (available_players.contains(i))
            available_teams += getPlayerTeam(i);

    /* Figure out current player */
    if (available_players.contains(::getPlayerId()))
        current_player = ::getPlayerId();
    else
        current_player = getNextPlayer(available_players, 0, 1);
    current_team = getPlayerTeam(current_player);

    /* Build the dialog */
    // VBox
    //   HBox
    //     VBox
    //       WBillDisplay
    //       WBillTotalDisplay
    //     UIScrollbar
    //   HBox
    //     "<"
    //     Race
    //     ">"
    //   HBox
    //     "Close"
    //     "Options"
    //     "Export"
    //     UISpacer
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1  = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& g11 = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& g2  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g3  = h.add(new UIGroup(UIHBoxLayout::instance5));

    bill_dpy  = &h.add(new WBillDisplay(1000, 20, true, fleet_cost));
    total_dpy = &h.add(new WBillTotalDisplay(fleet_cost));
    scrollbar = &h.add(new UIScrollbar(*bill_dpy));
    g1.add(g11);
    g11.add(*bill_dpy);
    g11.add(*total_dpy);
    g1.add(*scrollbar);
    add(g1);

    race_dpy = &h.add(new UIStaticText("", UIStaticText::StaticNormal, 0));
    race_dpy->setFlexible(true);
    g2.add(h.add(new UIButton("<", SDLK_TAB + ss_Shift)));
    g2.add(*race_dpy);
    g2.add(h.add(new UIButton(">", SDLK_TAB)));
    add(g2);

    UIButton& btn_export = h.add(new UIButton(2, UIKeyString(_("E - Export"))));
    btn_export.sig_fired.add(bill_dpy, &WBillDisplay::doExport);

    g3.add(h.add(new UIButton(1, _("Close"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(_("O - Options"), 'o')));
    g3.add(btn_export);
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(_("Help"), 'h')));
    add(g3);
    add(h.add(new WHelpWidget("pcc2:fleetcost")));
    add(h.add(new UIQuit(1)));

    pack();

    /* Build content */
    generateSummary();
}

void
WFleetCostDialog::browse(int delta)
{
    if (cost_opts.team) {
        current_team = getNextPlayer(available_teams, current_team, delta);
        if (available_players.contains(getPlayerId()) && getPlayerTeam(getPlayerId()) == current_team) {
            current_player = getPlayerId();
        } else {
            for (int i = 1; i <= NUM_PLAYERS; ++i) {
                if (available_players.contains(i) && getPlayerTeam(i) == current_team) {
                    current_player = i;
                    break;
                }
            }
        }
    } else {
        current_player = getNextPlayer(available_players, current_player, delta);
        current_team = getPlayerTeam(current_player);
    }
}

bool
WFleetCostDialog::isVisibleOwner(int pl) const
{
    if (cost_opts.team) {
        return getPlayerTeam(pl) == current_team;
    } else {
        return pl == current_player;
    }
}

void
WFleetCostDialog::generateSummary()
{
    /* Forget everything */
    fleet_cost.clear();

    /* Ships */
    int hullTech = 1;
    int beamTech = 1;
    int torpTech = 1;
    int engTech = 1;
    for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i) {
        const GSimShip& sh = state.getShip(i);
        if (isVisibleOwner(sh.getOwner())) {
            GCost cost;
            /* Hull, engines */
            if (sh.getHull() != 0) {
                cost += getHull(sh.getHull()).getCost();
                if (cost_opts.engine) {
                    cost += getEngine(sh.getEngineType()).getCost() * getHull(sh.getHull()).getNumEngines();
                }
                addTech(cost, sh.getOwner(), hullTech, getHull(sh.getHull()).getTech());
            }
            addTech(cost, sh.getOwner(), engTech, getEngine(sh.getEngineType()).getTech());
            /* Beams */
            if (sh.getBeamType() != 0) {
                cost += getBeam(sh.getBeamType()).getCost() * sh.getBeamCount();
                addTech(cost, sh.getOwner(), beamTech, getBeam(sh.getBeamType()).getTech());
            }
            /* Torps/Fighters */
            if (sh.getTorpType() != 0) {
                cost += getTorp(sh.getTorpType()).getCost() * sh.getTorpLauncherCount();
                if (cost_opts.torps) {
                    cost += getTorp(sh.getTorpType()).getTorpedo().getCost() * sh.getAmmo();
                }
                addTech(cost, sh.getOwner(), torpTech, getTorp(sh.getTorpType()).getTech());
            } else if (sh.getBayCount() != 0) {
                addFighters(cost, sh.getOwner(), sh.getAmmo());
            }
            fleet_cost.push_back(new WBillItem(0, getShipName(sh), cost, 1));
        }
    }

    /* Planet */
    if (state.hasPlanet() && isVisibleOwner(state.getPlanet().getOwner())) {
        const GSimPlanet& pl = state.getPlanet();
        if (cost_opts.planet && pl.getDefense() != 0) {
            GCost cost;
            cost.set(el_Money, 10*pl.getDefense());
            cost.set(el_Supplies, pl.getDefense());
            fleet_cost.push_back(new WBillItem(0, _("Planet"), cost, 1));
        }
        if (pl.hasBase()) {
            GCost cost;
            if (cost_opts.base) {
                cost = config.StarbaseCost(pl.getOwner());
                addFighters(cost, pl.getOwner(), pl.getBaseFighters());
                cost.add(el_Money, 10*pl.getBaseDefense());
                cost.add(el_Duranium, pl.getBaseDefense());
            }
            if (cost_opts.tech) {
                cost.add(el_Money, getBaseTechCost(pl.getOwner(), 1, pl.getBaseBeamTech()));
            }
            if (config.PlanetsHaveTubes() && WSimScreen::sim_opts.vcr_mode != GSimOptions::VcrHost && WSimScreen::sim_opts.vcr_mode != GSimOptions::VcrNuHost) {
                if (cost_opts.tech) {
                    cost.add(el_Money, getBaseTechCost(pl.getOwner(), 1, pl.getBaseTorpTech()));
                }
                if (cost_opts.torps && cost_opts.base) {
                    for (int i = 1; i <= NUM_TORPS; ++i) {
                        cost += getTorp(i).getTorpedo().getCost() * pl.getBaseTorps(i);
                    }
                }
            }
            if (!cost.isEmpty()) {
                fleet_cost.push_back(new WBillItem(0, _("Starbase"), cost, 1));
            }
        }
    }

    /* Display it */
    bill_dpy->update();
    total_dpy->update();
    scrollbar->drawWidget();  // FIXME: should actually update buttons
    if (cost_opts.team) {
        race_dpy->setText(getTeamName(current_team));
    } else {
        race_dpy->setText(player_racenames.getShortName(current_player));
    }
}

bool
WFleetCostDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_LEFT:
         case SDLK_TAB + ss_Shift:
            browse(-1);
            generateSummary();
            return true;

         case SDLK_RIGHT:
         case SDLK_TAB:
            browse(+1);
            generateSummary();
            return true;

         case SDLK_ESCAPE:
         case SDLK_RETURN:
            stop(1);
            return true;

         case 'o':
         case 'o' + ss_Ctrl:
            if (editFleetCostOptions(cost_opts)) {
                generateSummary();
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}
