/**
  *  \file client/dialogs/plugindlg.cc
  *  \brief Plugin Manager Dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/plugindlg.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "client/widgets/pluginlist.h"
#include "gfx/palette.h"
#include "ui/widgets/button.h"
#include "ui/widgets/statictext.h"
#include "ui/richdocument.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"
#include "util/pluginstall.h"
#include "arch/platform.h"
#include "util/progress.h"
#include "util/except.h"
#include "client/widgets/helpwidget.h"
#include "resmgr/resmgr.h"
#include "ui/richparser.h"
#include "ui/dialogs/filedlg.h"
#include "ui/richtext.h"
#include "util/unicodechars.h"
#include "io/glob.h"
#include "ui/commands.h"
#include "ui/widgets/quit.h"
#include "util/helpindex.h"

namespace {
    class WPluginDialog : public UIWindow {
     public:
        WPluginDialog(void loadPlugin(Plugin&));

        bool wantRestart() const
            { return restart; }

     private:
        ObjectHolder h;
        WPluginList* pList;
        UIRichDocument* pDoc;
        bool restart;
        void (*loadPlugin)(Plugin&);

        void init();
        void onAdd();
        void onRemove();
        void onClose();
        void onMove();
    };

    /** Render feature set into a RichDocument.
        Renders a whole paragraph.
        \param doc [out] Document
        \param fs [in] Feature set
        \param miss [in] Feature set of features that are actually missing
        \param ttl [in] Title of this feature set
        \param except [in] Do not render this feature */
    void addFeatures(RichDocument& doc,
                     const Plugin::FeatureSet& fs,
                     const Plugin::FeatureSet& miss,
                     const char* ttl, const string_t& except)
    {
        bool did = false;
        for (Plugin::FeatureSet::const_iterator i = fs.begin(); i != fs.end(); ++i) {
            if (i->first != except) {
                if (!did) {
                    doc.add(RichText(ttl).withAttribute(new RichTextStyleAttribute(RichTextStyleAttribute::Bold)));
                } else {
                    doc.add(", ");
                }

                RichText tmp(i->first);
                if (!i->second.empty()) {
                    tmp += " ";
                    tmp += i->second;
                }
                if (miss.find(i->first) != miss.end()) {
                    tmp.withAttribute(new RichTextColorAttribute(UIColor::tc_Red));
                    tmp.withAttribute(new RichTextStyleAttribute(RichTextStyleAttribute::Underline));
                }
                doc.add(tmp);
                did = true;
            }
        }
        if (did) {
            doc.addParagraph();
        }
    }

    /** Create path. We sweep problems under the carpet;
        actual usage of the created path will fail and be reported later on. */
    void
    createPath(string_t name)
    {
        string_t parent = Directory::getDirName(name);
        string_t child  = Directory::getFileName(name);

        try {
            Ptr<Directory> d = Directory::openDir(parent);
            Ptr<DirectoryEntry> e = d->getEntryByName(child);
            if (e->getFileType() != DirectoryEntry::FT_DIRECTORY)
                e->createAsDirectory();
        }
        catch (FileProblemException& /*e*/) {
            // ignore
        }
    }

    /** Check plugin preconditions. */
    static bool
    checkPreconditions(PluginManager& mgr, Plugin& plug)
    {
        // Check for conflicts
        std::vector<Plugin*> plugList;
        mgr.enumConflictingPlugins(plug, plugList);
        if (!plugList.empty()) {
            RichText text((format(_("Plugin \"%s\" cannot be installed.\n"), plug.getName())));

            string_t expl = "\n";
            expl += _("This plugin conflicts with the following plugins:");
            for (size_t i = 0, n = plugList.size(); i < n; ++i) {
                expl += format("\n" UTF_BULLET " %s (%s)", plugList[i]->getName(), plugList[i]->getId());
            }
            text += RichText(expl).withAttribute(RichTextStyleAttribute::Small);
            (UIMessageBox(text, _("Install Plugin"))).doOkDialog();
            return false;
        }

        // Check for preconditions
        Plugin::FeatureSet fset;
        mgr.enumFeatures(fset);
        if (!plug.isSatisfied(fset)) {
            RichText text((format(_("Plugin \"%s\" cannot be installed.\n"), plug.getName())));

            string_t expl = "\n";
            expl += _("This plugin requires the following features:");

            Plugin::FeatureSet missing;
            plug.enumMissingFeatures(fset, missing);
            for (Plugin::FeatureSet::const_iterator it = missing.begin(), e = missing.end(); it != e; ++it) {
                expl += "\n" UTF_BULLET " ";
                expl += it->first;
                string_t txt = it->first;
                if (!it->second.empty()) {
                    expl += " ";
                    expl += it->second;
                }
            }
            (UIMessageBox(text, _("Install Plugin"))).doOkDialog();
            return false;
        }
        return true;
    }
}

WPluginDialog::WPluginDialog(void loadPlugin(Plugin&))
    : UIWindow(_("Plugins"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      restart(false),
      loadPlugin(loadPlugin)
{
    init();
}

/** Initialize. Build the dialog. */
void
WPluginDialog::init()
{
    // VBox
    //   HBox
    //     VBox
    //       "Plugins:"
    //       HBox
    //         WPluginList
    //         UIScrollbar
    //     VBox
    //       WPluginInfo
    //   HBox
    //     "Add"
    //     "Delete"
    //     UISpacer
    //     "Close"

    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g11  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g111 = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& g12  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g2   = h.add(new UIGroup(UIHBoxLayout::instance5));

    pList = &h.add(new WPluginList(1000));
    g11.add(h.add(new UIStaticText(_("Plugins:"), UIStaticText::StaticTitle)));
    g111.add(*pList);
    g111.add(h.add(new UIScrollbar(*pList)));
    g11.add(g111);
    pDoc = &h.add(new UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(25, 10), 0));
    g12.add(*pDoc);
    g1.add(g11);
    g1.add(g12);
    add(g1);

    g2.add(h.add(new UIButton(1, _("Add"),    SDLK_INSERT)).addCall(this, &WPluginDialog::onAdd));
    g2.add(h.add(new UIButton(2, _("Remove"), SDLK_DELETE)).addCall(this, &WPluginDialog::onRemove));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(_("Help"),  'h')));
    g2.add(h.add(new UIButton(4, _("Close"),  SDLK_ESCAPE)).addCall(this, &WPluginDialog::onClose));
    add(g2);
    add(h.add(new WHelpWidget("pcc2:plugins")));
    add(h.add(new UIQuit(1 /* xref: onClose */)));

    pList->setFocusedWidget();
    pList->init(PluginManager::getInstance());
    pList->sig_moved.add(this, &WPluginDialog::onMove);

    pack();
    onMove();
}

/** "Add" action. */
void
WPluginDialog::onAdd()
{
    // Ask for file
    UIFileChooserDialog fc(_("Install Plugin"), Directory::openDir(Directory::getWorkingDirectory()),
                           FileNamePattern::getAllFilesWithExtensionPattern("c2p"));
    fc.addPattern(FileNamePattern::getAllFilesWithExtensionPattern("q"));
    fc.addPattern(FileNamePattern::getAllFilesWithExtensionPattern("res"));
    fc.addPattern(FileNamePattern::getAllFilesWithExtensionPattern("zip"));
    fc.addPattern(FileNamePattern::getAllFilesWithExtensionPattern("c2z"));
    fc.rescan();
    ui_root->centerWidget(fc);
    if (!ui_root->runModal(fc)) {
        return;
    }

    PluginManager& mgr = PluginManager::getInstance();
    try {
        // Create plugin directory
        string_t profileDirectory = Platform::getUserSettingsDirectory(1);
        string_t pluginDirName = Directory::makeFileName(profileDirectory, "plugins");
        createPath(profileDirectory);
        createPath(pluginDirName);

        // Create plugin installer
        Ptr<Directory> dir = Directory::openDir(pluginDirName);
        ProgressMonitor pm;
        PluginInstaller installer(mgr, *dir, pm);

        // Load the plugin
        Plugin* plug = installer.prepareInstall(fc.getResultFile());
        if (!plug) {
            (UIMessageBox(RichParser::textFromXml(_("File cannot be installed.\n"
                                                    "<small>The file you have chosen cannot be installed as a plugin. "
                                                    "A plugin is normally specified by a <tt>*.c2p</tt> file.</small>")),
                          _("Install Plugin"))).doOkDialog();
            return;
        }

        // Check ambiguities
        string_t altName;
        switch (installer.checkInstallAmbiguity(altName)) {
         case PluginInstaller::NoPlugin:
            break;

         case PluginInstaller::OnePlugin: {
            RichText txt(_("The file you have selected is not a plugin file."));
            txt += "\n\n";
            txt += RichText(format(_("It can be converted into a plugin. "
                                     "However, there is a plugin definition file (%s) next to it. "
                                     "If %s is part of that, it is strongly recommended to install the *.c2p file instead."),
                                   altName,
                                   Directory::getFileName(fc.getResultFile())))
                .withAttribute(RichTextStyleAttribute::Small);
            txt += "\n\n";
            txt += _("Do you want to proceed with the original file, or use the *.c2p file?");

            UIMessageBox box(txt, _("Install Plugin"));
            UIKeyString y(_("Proceed"));
            UIKeyString u(_("Use *.c2p"));
            box.addButton(cm_Enter, y.getString(), y.getKey());
            box.addButton(cm_No,    u.getString(), u.getKey());
            box.addButton(cm_Escape, _("Cancel"), SDLK_ESCAPE);
            box.addKey(cm_Enter, ' ');
            switch (box.run()) {
             case cm_Enter:
                // ok
                break;

             case cm_No:
                // use alternative
                plug = installer.prepareInstall(Directory::makeFileName(Directory::getDirName(fc.getResultFile()), altName));
                if (!plug) {
                    // oh crap, didn't work
                    (UIMessageBox(RichParser::textFromXml(_("File cannot be installed.\n"
                                                            "<small>There was trouble processing the alternative file.</small>")),
                                  _("Install Plugin"))).doOkDialog();
                    return;
                }
                break;

             case cm_Escape:
                // abort
                return;
            }
            break;
         }

         case PluginInstaller::MultiplePlugins: {
            RichText txt(_("The file you have selected is not a plugin file."));
            txt += "\n\n";
            txt += RichText(format(_("It can be converted into a plugin. "
                                     "However, there are multiple plugin definition files (*.c2p) next to it. "
                                     "If %s is part of one of these, it is strongly recommended to install the *.c2p file instead."),
                                   Directory::getFileName(fc.getResultFile())))
                .withAttribute(RichTextStyleAttribute::Small);
            txt += "\n\n";
            txt += _("Do you want to proceed anyway?");

            UIMessageBox box(txt, _("Install Plugin"));
            UIKeyString y(_("Proceed"));
            box.addButton(cm_Enter, y.getString(), y.getKey());
            box.addButton(cm_Escape, _("Cancel"), SDLK_ESCAPE);
            box.addKey(cm_Enter, ' ');
            if (box.run() != cm_Enter) {
                return;
            }
            break;
         }
        }
            

        // Check it
        if (!checkPreconditions(mgr, *plug)) {
            return;
        }

        // Confirm
        bool isUpdate = mgr.getPluginById(plug->getId()) != 0;
        {
            const char* msg = isUpdate
                ? _("Do you want to update plugin \"%s\" (%s)?)")
                : _("Do you want to install plugin \"%s\" (%s)?)");
            RichText txt(format(msg, plug->getName(), plug->getId()));
            if (plug->getDescription().size() > 0) {
                txt += "\n\n";
                txt += RichText(plug->getDescription()).withAttribute(RichTextStyleAttribute::Small);
            }
            if (!UIMessageBox(txt, _("Install Plugin")).doYesNoDialog()) {
                return;
            }
        }

        // If this is an update, unload the existing plugin
        if (isUpdate) {
            ResManager::getInstance().unloadByOrigin(plug->getId());
            HelpIndex::instance.removeFilesByOrigin(plug->getId());
            restart = true;
        }

        // Install the plugin
        installer.doInstall(false);
        if (loadPlugin) {
            loadPlugin(*plug);
        }
    }
    catch (FileProblemException& e) {
        messageBox(format("%s: %s", e.getFileName(), e.what()) + "\n" + _("Plugin not installed."),
                   _("Install Plugin"));
    }
    catch (std::exception& e) {
        messageBox(string_t(e.what()) + "\n" + _("Plugin not installed."),
                   _("Install Plugin"));
    }

    // Rebuild list
    pList->init(mgr);
}

/** "Remove" action. */
void
WPluginDialog::onRemove()
{
    PluginManager& mgr = PluginManager::getInstance();
    if (Plugin* plug = mgr.getPluginById(pList->getCurrentPluginId())) {
        // Check preconditions
        std::vector<Plugin*> tmp;
        mgr.enumDependingPlugins(*plug, tmp);
        if (!tmp.empty()) {
            RichText msg(_("This plugin is required by the following plugins:"));
            msg += "\n";
            for (size_t i = 0, n = tmp.size(); i < n; ++i) {
                msg += format("  %s (%s)\n", tmp[i]->getId(), tmp[i]->getName());
            }
            msg += _("It cannot be uninstalled.");
            UIMessageBox(msg, _("Uninstall Plugin")).doOkDialog();
            return;
        }

        // Ask user
        if (!UIMessageBox(format(_("Do you want to remove plugin '%s'?"), plug->getId()), _("Uninstall Plugin")).doYesNoDialog()) {
            return;
        }

        // Do it
        try {
            // Remove resources
            ResManager::getInstance().unloadByOrigin(plug->getId());
            HelpIndex::instance.removeFilesByOrigin(plug->getId());

            // Uninstall plugin
            ProgressMonitor pm;
            Ptr<Directory> dir = Directory::openDir(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "plugins"));
            if (!PluginInstaller(mgr, *dir, pm).doRemove(plug, false)) {
                messageBox(_("The plugin could not be uninstalled completely."),
                           _("Uninstall Plugin"));
            }
            restart = true;
        }
        catch (FileProblemException& e) {
            messageBox(format("%s: %s", e.getFileName(), e.what()) + "\n" + _("Plugin not uninstalled."),
                       _("Uninstall Plugin"));
        }
        catch (std::exception& e) {
            messageBox(string_t(e.what()) + "\n" + _("Plugin not uninstalled."),
                       _("Uninstall Plugin"));
        }

        // Rebuild list
        pList->init(mgr);
    }
}

/** "Close" action. */
void
WPluginDialog::onClose()
{
    stop(1);
}

/** Movement within the list.
    Redisplays the information. */
void
WPluginDialog::onMove()
{
    RichDocument& doc = pDoc->getDocument();
    doc.clear();
    doc.setPageWidth(pDoc->getExtent().w);

    if (Plugin* plug = PluginManager::getInstance().getPluginById(pList->getCurrentPluginId())) {
        // Header
        doc.add(RichText(plug->getName()).withAttribute(RichTextStyleAttribute::Big));
        doc.addNewline();
        doc.add(RichText(plug->isLoaded() ? UIColor::tc_Faded : UIColor::tc_Red,
                         format("(%d, %d)", plug->getId(), plug->isLoaded() ? _("loaded") : _("not loaded"))));
        doc.addParagraph();
        if (plug->getDescription().empty()) {
            doc.add(_("(no description given)"));
        } else {
            string_t desc = plug->getDescription();
            string_t::size_type n;
            while ((n = desc.find('\n')) != string_t::npos) {
                doc.add(desc.substr(0, n));
                doc.addParagraph();
                desc.erase(0, n+1);
            }
            doc.add(desc);
        }
        doc.addParagraph();

        // File list
        const Plugin::ItemList& items = plug->getItems();
        bool did = false;
        for (size_t i = 0; i < items.size(); ++i) {
            if (items[i].type != Plugin::Command) {
                if (!did) {
                    doc.add(RichText(_("Files: ")).withAttribute(RichTextStyleAttribute::Bold));
                } else {
                    doc.add(", ");
                }
                doc.add(items[i].name);
                did = true;
            }
        }
        if (did) {
            doc.addParagraph();
        }

        // Requires
        Plugin::FeatureSet empty, need, all, miss, offer;
        PluginManager::getInstance().enumFeatures(all);
        plug->enumMissingFeatures(empty, need);
        plug->enumMissingFeatures(all, miss);
        addFeatures(doc, need, miss, _("Requires: "), string_t());

        // Provides
        plug->addProvidedFeaturesTo(offer);
        addFeatures(doc, offer, empty, _("Provides: "), plug->getId());
    }

    doc.finish();
    pDoc->drawWidget();
}

/** Plugin manager dialog. */
void
doPluginDialog(void loadPlugin(Plugin&))
{
    WPluginDialog dlg(loadPlugin);
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);

    if (dlg.wantRestart()) {
        messageBox(_("You should restart PCC2 for the changes to take effect."), _("Plugins"));
    }
}
