/**
  *  \file client/dialogs/simflakedit.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/simflakedit.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/checkbox.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusgroup.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/focusiterator.h"
#include "game/sim.h"
#include "ui/uiglobal.h"
#include "flak/flakconf.h"
#include "game/config.h"

namespace {
    class WRatingEditorDialog : public UIWindow {
     public:
        WRatingEditorDialog();
        ~WRatingEditorDialog();

        void load(const GSimShip& sh, const GSimOptions& opts);
        void save(GSimShip& sh);

     private:
        ObjectHolder m_del;
        UIDecimalSelector m_ratingInput;
        UIDecimalSelector m_compensationInput;
        UICheckbox m_autoCheckbox;

        void init();
        void addFrame(UIDecimalSelector& sel, string_t label, int id);

        void onCheckboxClick();
        void onNumberInput();

        int32_t m_defaultRating;
        int32_t m_defaultCompensation;
    };
}

WRatingEditorDialog::WRatingEditorDialog()
    : UIWindow(_("FLAK Unit Ratings"), BLUE_WINDOW, COLOR_GRAY),
      m_del(),
      m_ratingInput(2000, 1, 1000000, 100),
      m_compensationInput(3000, 0, 32000, 100),
      m_autoCheckbox(1000, _("Automatic"), 'a', 2),
      m_defaultRating(1),
      m_defaultCompensation(1)
{
    init();
}

WRatingEditorDialog::~WRatingEditorDialog()
{ }

void
WRatingEditorDialog::load(const GSimShip& sh, const GSimOptions& opts)
{
    m_defaultCompensation = sh.getDefaultFlakCompensation(flak_config, opts);
    m_defaultRating = sh.getDefaultFlakRating(flak_config, opts);

    if ((sh.getFlags() & GSimShip::fl_RatingOverride) != 0) {
        m_autoCheckbox.setValue(0);
        m_ratingInput.setValue(sh.getFlakRatingOverride());
        m_compensationInput.setValue(sh.getFlakCompensationOverride());
    } else {
        m_autoCheckbox.setValue(1);
        m_ratingInput.setValue(m_defaultRating);
        m_compensationInput.setValue(m_defaultCompensation);
    }
}

void
WRatingEditorDialog::save(GSimShip& sh)
{
    int32_t flags = sh.getFlags() & ~GSimShip::fl_RatingOverride;
    if (m_autoCheckbox.getValue() != 0) {
        sh.setFlakRatingOverride(0);
        sh.setFlakCompensationOverride(0);
    } else {
        flags |= GSimShip::fl_RatingOverride;
        sh.setFlakRatingOverride(m_ratingInput.getValue());
        sh.setFlakCompensationOverride(m_compensationInput.getValue());
    }
    sh.setFlags(flags);
}

void
WRatingEditorDialog::init()
{
    // UIWindow
    //   UICheckbox    1000
    //   UIFocusGroup  1001
    //     Label
    //     "-"            2001
    //     UIInputLine    2000
    //     "+"            2002
    //   UIFocusGroup [Label   "-"   UIInputLine "+"]
    //   HBox [Spacer "OK" "Cancel"]

    setLayout(UIVBoxLayout::instance5);

    add(m_autoCheckbox);
    addFrame(m_ratingInput, _("Targeting"),          1001);
    addFrame(m_compensationInput, _("Compensation"), 1002);

    UIGroup& g = m_del.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(m_del.add(new UISpacer()));
    g.add(m_del.add(new UIButton(1, _("OK"),     SDLK_RETURN)).addStop());
    g.add(m_del.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    add(g);

    add(m_del.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));
    m_autoCheckbox.setFocusedWidget();
    m_autoCheckbox.sig_changed.add(this, &WRatingEditorDialog::onCheckboxClick);

    pack();
}

void
WRatingEditorDialog::addFrame(UIDecimalSelector& sel, string_t label, int id)
{
    UIFocusGroup& g = m_del.add(new UIFocusGroup(UIHBoxLayout::instance5, id, 2));
    UIButton& btnDec = m_del.add(new UIButton(sel.getId()+1, "-", '-'));
    UIButton& btnInc = m_del.add(new UIButton(sel.getId()+2, "+", '+'));
    UIStaticText& st = m_del.add(new UIStaticText(label, UIStaticText::StaticTitle));

    g.add(st);
    g.add(m_del.add(new UISpacer()));
    g.add(btnDec);
    g.add(sel);
    g.add(btnInc);
    sel.setFixedWidth(6 * font_ems[FONT_TITLE]);
    sel.setFocusedWidget();
    add(g);

    btnDec.sig_fired.add(static_cast<UINumberSelector*>(&sel), &UINumberSelector::decrementStep);
    btnInc.sig_fired.add(static_cast<UINumberSelector*>(&sel), &UINumberSelector::incrementStep);
    sel.sig_changed.add(this, &WRatingEditorDialog::onNumberInput);
}

void
WRatingEditorDialog::onCheckboxClick()
{
    if (m_autoCheckbox.getValue() != 0) {
        m_compensationInput.setValue(m_defaultCompensation);
        m_ratingInput.setValue(m_defaultRating);

        // The above will trigger onNumberInput, which may call setValue(0).
        m_autoCheckbox.setValue(1);
    }
}

void
WRatingEditorDialog::onNumberInput()
{
    if (m_compensationInput.getValue() != m_defaultCompensation || m_ratingInput.getValue() != m_defaultRating) {
        m_autoCheckbox.setValue(0);
    }
}

void
editSimulatorFlakRatings(GSimShip& ship, const GSimOptions& opts)
{
    WRatingEditorDialog dlg;
    dlg.load(ship, opts);

    ui_root->centerWidget(dlg);
    int result = ui_root->runModal(dlg);

    if (result != 0) {
        dlg.save(ship);
    }
}
