/**
  *  \file client/dialogs/zoomlevel.cc
  *  \brief Zoom level dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/zoomlevel.h"
#include "client/chart.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"

namespace {
    class Dialog {
     public:
        Dialog(const GChartViewport& renderer, ZoomLevel& result)
            : m_renderer(renderer),
              m_result(result),
              m_input(20, 10, 0, 0, 0),
              m_ok(1, _("OK"), SDLK_RETURN),
              m_cancel(0, _("Cancel"), SDLK_ESCAPE)
            {
                m_input.setText(formatZoomLevel(renderer.getZoomMultiplier(), renderer.getZoomDivider()));
                m_input.sig_changed.add(this, &Dialog::onChange);
                m_ok.sig_fired.add(this, &Dialog::onOK);
                m_cancel.addStop();
                onChange();
            }

        bool run()
            {
                ObjectHolder h;
                UIWindow& win = h.add(new UIWindow(_("Zoom"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
                win.setLayout(UIVBoxLayout::instance5);
                win.add(h.add(new UIStaticText(_("Zoom level:"), UIStaticText::StaticTitle)));
                win.add(m_input);
                win.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("Enter zoom level in format \"3\" (zoom in) or \"1/4\" resp. \"1:4\" (zoom out)."), 20*font_ems[FONT_NORMAL])));

                UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
                g.add(h.add(new UISpacer()));
                g.add(m_ok);
                g.add(m_cancel);
                win.add(g);
                win.pack();

                ui_root->centerWidget(win);
                int n = ui_root->runModal(win);
                return n != 0;
            }

        void onOK()
            {
                int m = 0, d = 0;
                if (parseZoomLevel(m_input.getText(), m, d) && m_renderer.isValidZoomLevel(m, d)) {
                    m_result.mult = m;
                    m_result.divi = d;
                    ui_root->stop(1);
                }
            }

        void onChange()
            {
                m_ok.setState(UIWidget::st_Disabled, !isValid());
            }

        bool isValid()
            {
                int m = 0, d = 0;
                return parseZoomLevel(m_input.getText(), m, d) && m_renderer.isValidZoomLevel(m, d);
            }

     private:
        const GChartViewport& m_renderer;
        ZoomLevel& m_result;
        UIInputLine m_input;
        UIButton m_ok;
        UIButton m_cancel;
    };
}

bool
editZoomLevel(const GChartViewport& renderer, ZoomLevel& result)
{
    // ex chartdlg.pas:EditZoom
    return Dialog(renderer, result).run();
}
