/**
  *  \file client/tiles/basetaskcmd.cc
  *  \brief WBaseAutoTaskCommandTile tile
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/basetaskcmd.h"
#include "client/actions.h"
#include "client/dialogs/basetaskbuild.h"
#include "client/screens/planettaskscreen.h"
#include "client/tasksupport.h"
#include "client/widgets/buildorderdisplay.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/planettransfer.h"
#include "game/ship.h"
#include "game/tables.h"
#include "int/args.h"
#include "int/if/baseif.h"
#include "int/if/basepred.h"
#include "int/if/buildpred.h"
#include "int/if/ifutil.h"
#include "int/predict.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/strtab.h"
#include "util/unicodechars.h"
#include "ui/widgets/decimalsel.h"

static const char*const buttons[] = {
    N_("1 - Orders"),
    N_("2 - Cargo"),
    N_("3 - Misc."),
};

namespace {
    void addMissingCargo(string_t& out, const char* lbl, int32_t miss)
    {
        if (miss > 0) {
            if (!out.empty()) {
                out += ' ';
            }
            out += numToString(miss);
            out += lbl;
        }
    }
}


/************************ WBaseAutoTaskCommandTile ***********************/

WBaseAutoTaskCommandTile::WBaseAutoTaskCommandTile(Personality pers, GObjectSelection& selection)
    /* We actually need 5 lines; there's room for 7, so we take
       those for safety in case something wraps. */
    : WTaskMessageTile(GfxPoint(315, 7*font_heights[FONT_NORMAL])),
      atsel(dynamic_cast<WPlanetAutoTaskSelection*>(&selection)),
      pers(pers)
{
    init();
}

WBaseAutoTaskCommandTile::~WBaseAutoTaskCommandTile()
{ }

void
WBaseAutoTaskCommandTile::init()
{
    // Controls
    setLayout(RightJustified, 0, true);
    StaticTranslatedStringTable tab(buttons, countof(buttons));
    const int buttonWidth = getTextWidth(*fonts[FONT_SMALL], tab) + 20;
    const int height = font_heights[FONT_SMALL] * 3/2;
    for (uint_t i = 0; i < countof(buttons); ++i) {
        UIButton& btn = h.add(new UIButton(1000+i, tab.get(i), '1'+i));
        btn.addCommand();
        btn.setState(UIButton::bf_LeftJust + UIButton::bf_SmallSize, true);
        addControl(GfxRect(0, (height+2)*i, buttonWidth, height), btn);
    }

    btn_edit = &h.add(new UIButton("E", 'e'));
    if (pers == BasePersonality) {
        addControl(GfxRect(buttonWidth-20, 7*font_heights[FONT_NORMAL]-20, 20, 20), *btn_edit);
        widthReserve = 5 + buttonWidth;
    } else {
        widthReserve = 0;
    }

    initTask(atsel, buttonWidth);
}

void
WBaseAutoTaskCommandTile::drawData(GfxCanvas& can)
{
    // Draw task
    if (drawTask(can)) {
        // This means the task drew it stuff to the left of buttonWidth (=5+buttonWidth).
        // We must still neutralize the "Edit" text which may write across buttonWidth.
        GfxRect area = getInnerSize(0, 0);
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(getSkin());
        drawBackground(ctx, GfxRect(area.x + area.w - widthReserve,
                                    area.y + area.h - 20,
                                    widthReserve - 25,
                                    20));
        return;
    }

    // Sanity check
    if (atsel == 0) {
        return;
    }

    // Clear
    GfxRect area = getInnerSize(5, widthReserve);
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, area);

    // Label
    if (pers == BasePersonality) {
        ctx.setTextAlign(2, 1);
        ctx.setColor(UIColor::tc_Static);
        outText(ctx, getExtent().w - 25, area.h - 10, _("Edit"));
        ctx.setTextAlign(0, 0);
    }

    // Parse current command
    std::auto_ptr<WBuildCommandParser::Result> data;
    if (GPlanet* pl = dynamic_cast<GPlanet*>(atsel->getCurrentObject())) {
        WBuildCommandParser parser(*pl);
        if (IntExecutionContext* proc = atsel->getTask(false)) {
            parser.loadLimit(atsel->getEditor(), *proc, atsel->getCursor());
        }
        parser.predictStatement(atsel->getEditor(), atsel->getCursor());
        data = parser.getResult();
    }

    if (data.get() != 0) {
        // It's a valid build order, display it
        int lineHeight = font_heights[FONT_NORMAL];
        ctx.setColor(UIColor::tc_Contrast);
        outTextF(ctx, area.splitY(lineHeight), _("Build Order:"));

        ctx.setColor(UIColor::tc_Static);
        string_t fmt = data->info;
        string_t::size_type n;
        while ((n = fmt.find('\n')) != string_t::npos) {
            outTextF(ctx, area.splitY(lineHeight), fmt.substr(0, n));
            fmt.erase(0, n+1);
        }
        if (!fmt.empty()) {
            outTextF(ctx, area.splitY(lineHeight), fmt);
        }

        // Validate cost
        string_t missing;
        addMissingCargo(missing, "T",   data->missingAmount.get(el_Tritanium));
        addMissingCargo(missing, "D",   data->missingAmount.get(el_Duranium));
        addMissingCargo(missing, "M",   data->missingAmount.get(el_Molybdenum));
        addMissingCargo(missing, "mc",  data->missingAmount.get(el_Money));
        addMissingCargo(missing, "sup", data->missingAmount.get(el_Supplies));

        if (!missing.empty()) {
            ctx.setColor(UIColor::tc_Yellow);
            outTextF(ctx, area.splitY(lineHeight), _("Additional resources needed:"));
            outTextF(ctx, area.splitY(lineHeight), UTF_BULLET " " + missing);
        } else {
            outTextF(ctx, area.splitY(lineHeight), _("Sufficient resources available"));
        }

        btn_edit->setState(UIButton::st_Disabled, data->type != WBuildCommandParser::OrderType_Ship);
    } else {
        // Not valid. Disable button.
        btn_edit->setState(UIButton::st_Disabled, true);
    }
}

void
WBaseAutoTaskCommandTile::createSetMissionCommand()
{
    // Find out old mission
    GPlanet* pl = dynamic_cast<GPlanet*>(atsel->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    int msn = pred.getPlanet().getBaseMission();
    if (chooseBaseMission(msn, "pcc2:basetaskscreen")) {
        atsel->insert(format("SetMission %d   %% %s", msn, getBaseMissionName(msn)));
    }
}

void
WBaseAutoTaskCommandTile::createEnqueueShipCommand()
{
    // Find out preconditions
    GPlanet* pl = dynamic_cast<GPlanet*>(atsel->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    // We'll be making an EnqueueShip command, which will wait until its
    // preceding command is performed. So do that now.
    if (pred.getPlanet().getBaseBuildOrderHullSlot() > 0) {
        pred.advanceTurn();
    }

    // What can we build?
    std::vector<int> hulls;
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
        if (int h = getTruehull(pl->getOwner(), i)) {
            hulls.push_back(h);
        }
    }

    // Build dialog
    WBaseTaskShipBuildDialog dialog(pred.getPlanet(), hulls);
    ui_root->centerWidget(dialog);
    if (dialog.runDialog()) {
        atsel->insert(makeBuildOrderCommand("EnqueueShip", dialog.getBuildOrder()));
    }
}

void
WBaseAutoTaskCommandTile::createBuildStructuresCommand(const char* title, const char* verb)
{
    UIDecimalSelector select(0, 0, 1000, 10);
    if (select.doStandardDialog(title, title, pers==BasePersonality ? "pcc2:basetaskscreen" : "pcc2:planettaskscreen")) {
        atsel->insert(format("%s %s", verb, select.getValue()));
    }
}

void
WBaseAutoTaskCommandTile::editCurrentCommand()
{
    // Are we looking at a build command?
    WBaseTaskBuildCommandParser parser;
    parser.predictStatement(atsel->getEditor(), atsel->getCursor());
    if (!parser.getOrder().hull) {
        return;
    }

    // Find out preconditions
    GPlanet* pl = dynamic_cast<GPlanet*>(atsel->getCurrentObject());
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(atsel->getEditor(), atsel->getCursor());

    // We'll be making an EnqueueShip command, which will wait until its
    // preceding command is performed. So do that now.
    if (pred.getPlanet().getBaseBuildOrderHullSlot() > 0) {
        pred.advanceTurn();
    }

    // What can we build?
    std::vector<int> hulls;
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
        if (int h = getTruehull(pl->getOwner(), i)) {
            hulls.push_back(h);
        }
    }
    pred.getPlanet().setBaseBuildOrder(parser.getOrder());

    // Build dialog
    WBaseTaskShipBuildDialog dialog(pred.getPlanet(), hulls);
    ui_root->centerWidget(dialog);
    if (dialog.runDialog()) {
        string_t cmd = parser.getVerb();
        if (cmd == "ENQUEUESHIP") {
            cmd = "EnqueueShip";
        }
        if (cmd == "BUILDSHIP") {
            cmd = "BuildShip";
        }
        atsel->replace(atsel->getCursor(), 1, makeBuildOrderCommand(cmd, dialog.getBuildOrder()));
    }
}

bool
WBaseAutoTaskCommandTile::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case 1000:
        if (atsel != 0) {
            StringList sl;
            if (pers == BasePersonality) {
                sl.add(1, _("Enqueue Ship Build Order"));
                sl.add(2, _("Set Mission"));
                sl.add(3, _("Set FCode"));
            } else {
                sl.add(3, _("Set FCode"));
                sl.add(4, _("Build starbase"));
                sl.add(5, _("Build factories"));
                sl.add(6, _("Build mines"));
                sl.add(7, _("Build defense posts"));
            }
            UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
            if (list.doStandardMenu(sender)) {
                switch (list.getKey()) {
                 case 1:
                    createEnqueueShipCommand();
                    break;
                 case 2:
                    createSetMissionCommand();
                    break;
                 case 3:
                    atsel->createSetFCodeCommand();
                    break;
                 case 4:
                    atsel->insert("BuildBaseWait");
                    break;
                 case 5:
                    createBuildStructuresCommand(_("Build factories"), "BuildFactoriesWait");
                    break;
                 case 6:
                    createBuildStructuresCommand(_("Build mines"), "BuildMinesWait");
                    break;
                 case 7:
                    createBuildStructuresCommand(_("Build defense posts"), "BuildDefenseWait");
                    break;
                }
            }
        }
        return true;
     case 1001:
        if (atsel != 0) {
            atsel->createCargoCommand(sender);
        }
        return true;
     case 1002:
        if (atsel != 0) {
            atsel->createCommonCommand(sender, false);
        }
        return true;
    }
    return WTile::handleCommand(cmd, sender);
}

bool
WBaseAutoTaskCommandTile::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'e':
            if (atsel != 0 && pers == BasePersonality) {
                editCurrentCommand();
            }
            return true;
        }
    }
    return WTile::handleEvent(event, second_pass);
}
