/**
  *  \file client/tiles/shipcargo.cc
  *  \brief Ship Cargo Tile
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipcargo.h"

#include "cpluslib/format.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"            // drawFrameDown
#include "util/misc.h"          // clansToString
#include "util/nls.h"
#include "game/cmds.h"
#include "game/config.h"
#include "game/global.h"
#include "ui/widgets/button.h"

namespace {
    /** Ship cargo transfer status button. Displays a symbol when a cargo transfer
        is active, and generates an appropriate key event when the user clicks it. */
    class WShipCargoButton : public UIAbstractButton,
                             public GObjectSelectionObserver
    {
        string_t text;
        bool ship;

        void set(UIKey key, string_t text);
     public:
        WShipCargoButton(GObjectSelection& selection, bool ship);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);

        void onObjectChanged();
    };
}

/** Constructor.
    \param selection Object selection to observe
    \param ship true: observe ship/ship transporter. false: observe ship/planet transporter */
WShipCargoButton::WShipCargoButton(GObjectSelection& selection, bool ship)
    : UIAbstractButton(),
      GObjectSelectionObserver(selection),
      text(),
      ship(ship)
{
    onObjectChanged();
}

/** Set content.
    \param key Key to generate
    \param text Text to display */
void
WShipCargoButton::set(UIKey key, string_t text)
{
    setKey(key);
    if (text != this->text) {
        this->text = text;
        drawWidget();
    }
}

void
WShipCargoButton::drawContent(GfxCanvas& can)
{
    /* Right-justified, yellow */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setColor(UIColor::tc_Yellow);
    ctx.setTextAlign(2, 0);
    ctx.setSolidBackground();
    outTextF(ctx, getExtent().getRightX(), getExtent().y, getExtent().w, text);

    can.update(getExtent());
}

void
WShipCargoButton::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_SMALL].scaledBy(2, 1);
    info.flags = 0;
}

void
WShipCargoButton::onObjectChanged()
{
    GShip* s = dynamic_cast<GShip*>(getCurrentObject());
    if (!s || !s->isPlayable(GObject::Playable)) {
        set(0, string_t());
    } else if (ship) {
        /* We're observing the ship->ship transporter */
        if (s->isTransporterActive(GShip::t_Transfer)) {
            /* Ship transfer: "->S", Ctrl+S */
            set(ss_Ctrl + 's', "\xE2\x86\x92""S");
        } else {
            set(0, string_t());
        }
    } else {
        /* We're observing the ship->planet transporter */
        bool bum = config.AllowBeamUpMultiple()
            && getDisplayedTurn().getCommands(s->getOwner()).getCommand(GCommand::phc_Beamup, s->getId()) != 0;

        if (s->isTransporterActive(GShip::t_Unload)) {
            if (s->getTransporter(GShip::t_Unload).id == 0) {
                /* Jettison: "->J", Ctrl+J */
                set(ss_Ctrl + 'j', "\xE2\x86\x92""J");
            } else if (bum) {
                /* Beam up + down: "<->P", Ctrl+P */
                set(ss_Ctrl + 'p', "\xE2\x86\x94""P");
            } else {
                /* Beam down: "->P", Ctrl+P */
                set(ss_Ctrl + 'p', "\xE2\x86\x92""P");
            }
        } else {
            if (bum) {
                /* Beam-up only: "<-P", no hot-key */
                set(0, "\xE2\x86\x90""P");
            } else {
                set(0, string_t());
            }
        }
    }
}

/***************************** WShipCargoTile ****************************/

WShipCargoTile::WShipCargoTile(GObjectSelection& selection)
    // FIXME: magic numbers
    : WObjectSelectionObserverTile(GfxPoint(315, 4 * font_heights[FONT_NORMAL] + 14 + 3), selection)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(25, 0, 20, 20), h.add(new UIButton("C", 'c')));
    addControl(GfxRect(0,  0, 20, 20), h.add(new UIButton("U", 'u')));
    addControl(GfxRect(25, 25, 20, 15), h.add(new WShipCargoButton(selection, false)));
    addControl(GfxRect(25, 40, 20, 15), h.add(new WShipCargoButton(selection, true)));
}

void
WShipCargoTile::drawShipCargoTile(UIAbstractTableWriter& tw, const GShip& sh)
{
    // Column 1: ore names
    // Column 2: ore amounts
    // Column 3: space
    // Column 4: rest
    // Column 5: rest amounts
    // Column 6: units

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    const char* itemFormat = _("%d kt");
    bool fuelWarning = (sh.getCargo(el_Neutronium) == 0);
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth(format(itemFormat, numToString(9999))));
    tw.writeVariableCell(fuelWarning ? UIColor::tc_Red : UIColor::tc_Green, format(itemFormat, numToString(sh.getCargo(el_Neutronium))));
    tw.writeVariableCell(UIColor::tc_Green, format(itemFormat, numToString(sh.getCargo(el_Tritanium))));
    tw.writeVariableCell(UIColor::tc_Green, format(itemFormat, numToString(sh.getCargo(el_Duranium))));
    tw.writeVariableCell(UIColor::tc_Green, format(itemFormat, numToString(sh.getCargo(el_Molybdenum))));

    tw.newLColumn(5);           // spacer

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Colonists:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Total Mass:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999"));
    tw.writeVariableCell(UIColor::tc_Green, clansToString(sh.getCargo(el_Colonists)));
    tw.writeVariableCell(UIColor::tc_Green, numToString((sh.getCargo(el_Supplies))));
    tw.writeVariableCell(UIColor::tc_Green, numToString((sh.getCargo(el_Money))));
    tw.writeVariableCell(UIColor::tc_Green, numToString((sh.getMass())));

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, "");
    tw.writeFixedCell(UIColor::tc_Green, _("kt"));
    tw.writeFixedCell(UIColor::tc_Green, _("mc"));
    tw.writeFixedCell(UIColor::tc_Green, _("kt"));

    tw.finish();
}

void
WShipCargoTile::drawData(GfxCanvas& can)
{
    GShip* s = dynamic_cast<GShip*>(getCurrentObject());
    if (!s)
        return;

    GfxRect pos = getInnerSize(0, 0);
    const int y = font_heights[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 0), FONT_NORMAL);
    drawShipCargoTile(tw, *s);

    int cargo = s->getHull().getCargo();
    drawCargoBar(can, pos.x + 5,   pos.y + 4*y + 4, 130, s->getCargo(el_Neutronium), s->getHull().getFuel(), COLOR_GREENBLACK, _("Fuel: %d free"));
    drawCargoBar(can, pos.x + 145, pos.y + 4*y + 4, 130, cargo - s->getFreeCargo(), cargo, COLOR_BLUE, _("Cargo: %d free"));
}

void
WShipCargoTile::drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max,
                             int color, const char* fmt)
{
    drawFrameDown(can, GfxRect(x, y, wi, 10));
    ++x;
    ++y;
    wi -= 2;

    int split = max ? wi * have / max : 0;
    if (have > 0 && split == 0)
        split = 1;
    if (have >= max)
        split = wi;
    else if (split >= wi)
        --split;

    if (split)
        drawSolidBar(can, GfxRect(x, y, split, 8), standard_colors[color]);
    if (split != wi)
        drawSolidBar(can, GfxRect(x + split, y, wi-split, 8), standard_colors[COLOR_BLACK]);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setTextAlign(1, 0);
    if (have > max)
        ctx.setColor(COLOR_RED);
    else
        ctx.setColor(COLOR_WHITE);
    ::outTextF(ctx, x + wi/2, y-1, wi, format(fmt, numToString(max-have)));
}


/************************** WNarrowShipCargoTile *************************/


WNarrowShipCargoTile::WNarrowShipCargoTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(250, 4*font_heights[FONT_NORMAL]), selection)
{ }

void
WNarrowShipCargoTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());

    // Is it a ship?
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (!sh) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    // Draw it
    static const GCargoType elements[] = { el_Neutronium, el_Tritanium, el_Duranium, el_Molybdenum };
    static const GCargoType elements2[] = { el_Supplies, el_Colonists, el_Money };
    static const char*const units2[] = { N_("kt"), N_("cl."), N_("mc") };

    UITableWriter tw(ctx, getInnerSize(0, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    tw.newRColumn(4*font_ems[FONT_NORMAL]);
    for (unsigned int i = 0; i < countof(elements); ++i) {
        if (sh->getCargo(elements[i]).isKnown())
            tw.writeVariableCell(
                elements[i] == el_Neutronium && sh->getCargo(elements[i]) == 0 ? UIColor::tc_Red : UIColor::tc_Green,
                numToString(sh->getCargo(elements[i])));
        else
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    tw.newLColumn(2*font_ems[FONT_NORMAL]);
    for (unsigned int i = 0; i < countof(elements); ++i) {
        if (sh->getCargo(elements[i]).isKnown())
            tw.writeVariableCell(
                elements[i] == el_Neutronium && sh->getCargo(elements[i]) == 0 ? UIColor::tc_Red : UIColor::tc_Green,
                _("kt"));
        else
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Sup:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Col:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Cash:"));

    tw.newRColumn(4*font_ems[FONT_NORMAL]);
    for (unsigned int i = 0; i < countof(elements2); ++i) {
        if (sh->getCargo(elements2[i]).isKnown())
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh->getCargo(elements2[i])));
        else
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    tw.newLColumn();
    for (unsigned int i = 0; i < countof(elements2); ++i) {
        if (sh->getCargo(elements2[i]).isKnown())
            tw.writeVariableCell(UIColor::tc_Green, _(units2[i]));
        else
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }
}
