/**
  *  \file client/widgets/configedit.cc
  *  \brief WConfigEditor widget
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/configedit.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "util/conf.h"
#include "ui/widgets/inputline.h"

/***************************** WConfigEditor *****************************/

/** Constructor. */
WConfigEditor::WConfigEditor()
    : UIWidget(),
      btn_edit(0, _("Enter - Change"), SDLK_RETURN),
      name(),
      editor(0),
      arg()
{
    init();
}

WConfigEditor::~WConfigEditor()
{ }

/** Set option to edit.
    \param name Name to display
    \param editor Editor; can be null to disable this widget
    \param arg Argument to pass to editor */
void
WConfigEditor::set(string_t name, Editor* editor, string_t arg)
{
    this->name   = name;
    this->editor = editor;
    this->arg    = arg;
    btn_edit.setState(st_Disabled, editor == 0);
    drawWidget();
}

void
WConfigEditor::getLayoutInfo(LayoutInfo& li)
{
    LayoutInfo li_btn;
    btn_edit.getLayoutInfo(li_btn);

    int width  = std::max(li_btn.min_size.x, font_ems[FONT_NORMAL] * 20);
    int height = li_btn.min_size.y + 5 + font_heights[FONT_NORMAL] + font_heights[FONT_NORMAL_BOLD];

    li.pref_size = li.min_size = GfxPoint(width, height);
    li.flags = li_FlexibleH;
}

void
WConfigEditor::drawContent(GfxCanvas& can)
{
    GfxRect r(getExtent());
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();

    ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
    ctx.setColor(UIColor::tc_Heading);
    outTextF(ctx, r.splitY(font_heights[FONT_NORMAL_BOLD]), name);

    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    string_t text;
    if (editor != 0) {
        text = editor->toString(arg);
        if (text.empty()) {
            text = _("(empty)");
            ctx.setColor(UIColor::tc_Faded);
        }
    }
    outTextF(ctx, r.splitY(font_heights[FONT_NORMAL]), text);

    can.update(getExtent());
}

void
WConfigEditor::onResize(GfxRect oldSize, bool byLayout)
{
    UIWidget::onResize(oldSize, byLayout);

    LayoutInfo li_btn;
    btn_edit.getLayoutInfo(li_btn);

    GfxRect pos_btn(0, 0, li_btn.min_size);
    pos_btn.moveToEdge(getExtent(), 2, 2, 0);
    btn_edit.setExtent(pos_btn, true);
}

/** Initialisation. Sets up the widget. */
void
WConfigEditor::init()
{
    add(btn_edit, GfxRect(0, 0, 10, 10));
    btn_edit.setState(st_Disabled, true);
    btn_edit.addCall(this, &WConfigEditor::onEdit);
}

/** Editor. Called when user pushes the "Edit" button. */
void
WConfigEditor::onEdit()
{
    if (editor != 0) {
        editor->edit(arg);
        sig_change.raise();
        drawWidget();
    }
}

/************************** WConfigOptionEditor **************************/

WConfigOptionEditor::WConfigOptionEditor(Config& cfg)
    : cfg(cfg)
{ }

string_t
WConfigOptionEditor::toString(string_t arg)
{
    if (ConfigOption* opt = cfg.getConfigOption(arg)) {
        return opt->toString();
    } else {
        return string_t();
    }
}

void
WConfigOptionEditor::edit(string_t arg)
{
    if (ConfigOption* opt = cfg.getConfigOption(arg)) {
        string_t value = opt->toString();
        while (UIInputLine::run(value, 1000, _("Edit Option"), _("New Value:"), 20, 0)) {
            try {
                opt->setAndMarkUpdated(value, ConfigOption::srcUser);
                break;
            }
            catch (...) { }
        }
    }
}
