/**
  *  \file client/widgets/funcinfo.cc
  *  \brief Hull Function Information Widget
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/funcinfo.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/struct.h"
#include "ui/uiglobal.h"
#include "ui/skin.h"
#include "ui/util.h"
#include "game/config.h"
#include "resmgr/resmgr.h"
#include "gfx/palette.h"

namespace {
    class FramedImageObject : public RichDocumentBlockObject {
     public:
        FramedImageObject(Ptr<GfxPixmap> pix)
            : pix(pix)
            { }

        GfxPoint getSize()
            { return GfxPoint(pix->getWidth()+2, pix->getHeight()+2); }

        void draw(GfxContext& ctx, GfxRect area)
            {
                drawFrameDown(ctx.getCanvas(), area);
                area.grow(-1, -1);
                drawSolidBar(ctx.getCanvas(), area, standard_colors[COLOR_BLACK]);
                pix->blitSized(ctx.getCanvas(), area.x, area.y, area.w, area.h);
            }
     private:
        Ptr<GfxPixmap> pix;
    };
}

WHullFunctionInfo::WHullFunctionInfo(int owner, int damage, const GShipExperienceInfo& expInfo)
    : UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(36, 11), 0),
      owner(owner), damage(damage), expInfo(expInfo)
{ }

WHullFunctionInfo::~WHullFunctionInfo()
{ }

void
WHullFunctionInfo::setFunction(const GHullFunction* hf)
{
    RichDocument& doc = getDocument();
    doc.clear();
    if (hf != 0) {
        /* Function title */
        const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(hf->getBasicFunctionId());
        int damageLimit;
        string_t picName;
        if (ptr) {
            string_t picName = ptr->getPictureName();
            if (!picName.empty()) {
                Ptr<GfxPixmap> pix = requestPixmap(ResId("abilities.32." + picName));
                if (pix.ptr()) {
                    std::auto_ptr<RichDocumentBlockObject> obj(new FramedImageObject(pix));
                    doc.addFloatObject(obj, false);
                }
            }
            doc.add(RichText(ptr->getDescription()).withAttribute(RichTextStyleAttribute::Bold));
            damageLimit = ptr->getDamageLimit(owner, config);
        } else {
            doc.add(RichText(format(_("Hull Function #%d"), hf->getBasicFunctionId())).withAttribute(RichTextStyleAttribute::Bold));
            damageLimit = -1;
        }
        doc.addParagraph();

        /* Availability information */
        string_t tmp = formatPlayerSet(hf->getPlayers());
        if (tmp.size()) {
            doc.add(string_t(format(_("Available to %s"), tmp)));
            doc.addNewline();
        }
        tmp = formatExpLevelSet(hf->getLevels());
        if (tmp.size()) {
            int i = 0;
            while (i <= NUM_EXPERIENCE_LEVELS && !hf->getLevels().contains(i)) {
                ++i;
            }
            if (i > 0 && i <= NUM_EXPERIENCE_LEVELS) {
                int minExp = config.ExperienceLevels(i);
                int numTurns = getNumTurnsUntil(minExp, expInfo);
                if (numTurns > 0) {
                    tmp += format(_(" (%d EP; in %d turn%!1{s%} at +%d/turn)"), minExp, numTurns, expInfo.pointGrowth.isKnown() ? int(expInfo.pointGrowth) : 0);
                } else {
                    tmp += format(_(" (%d EP)"), minExp);
                }
            }
            doc.add(string_t(format(_("Available at %s"), tmp)));
            doc.addNewline();
        }

        if (hf->getKind() == GHullFunction::AssignedToRace) {
            if (hf->getPlayers().contains(GPlayerSet::allUpTo(NUM_PLAYERS) - 0)) {
                doc.add(_("Universal ability"));
            } else {
                doc.add(_("Racial ability"));
            }
        } else if (hf->getKind() == GHullFunction::AssignedToShip) {
            doc.add(_("Assigned to ship"));
        } else {
            doc.add(_("Assigned to ship class"));
        }

        if (damageLimit >= 0 && damageLimit < 100) {
            doc.addNewline();
            doc.add(string_t(format(_("Fails at %d%% damage"), damageLimit)));
            if (damage >= damageLimit) {
                doc.add(" ");
                doc.add(RichText(UIColor::tc_Red, _("(currently broken)")));
            }
        }

        doc.addParagraph();

        /* Description */
        if (ptr) {
            doc.add(ptr->getExplanation());
        }
    }
    doc.finish();
    handleDocumentUpdate();
}
