/**
  *  \file flak/vis-iso.cc
  *  \brief ISO-3D visualizer
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "flak/vis-sprites.h"
#include "flak/vis.h"
#include "game/formula.h"
#include "game/spec.h"
#include "gfx/blit.h"
#include "gfx/palette.h"
#include "gfx/sprite3d.h"
#include "resmgr/resid.h"
#include "resmgr/resmgr.h"

namespace {
    enum {
        SHIP_RADIUS = 5
    };

    //! Isometric Ship Sprite.
    class Flak3DShipSprite : public GfxSprite3D {
        Flak3DVisualizer& vis;
        FlakBattle::ship_t ship;
        int angle;
        int timer;
        Ptr<GfxPixmap> pix;
     public:
        Flak3DShipSprite(Flak3DVisualizer& vis,
                         FlakBattle::ship_t ship)
            : GfxSprite3D(FlakIsoVisualizer::layer_Ships),
              vis(vis), ship(ship), angle(-1), timer(0)
            { tick(); }
        void draw(GfxCanvas& can);
        void tick();
    };

    double wrapAngle(double r)
    {
        if (r < -PI) {
            return r + 2*PI;
        } else if (r > PI) {
            return r - 2*PI;
        } else {
            return r;
        }
    }
}

static const int num_angles     = 16;
static const int min_turn_delay = 5;

/** Get rotated ship picture.
    \param hull hull number
    \param angle angle, [0, num_angles). */
static Ptr<GfxPixmap>
getShipPicture(int hull, int angle)
{
    ResManager& mgr = ResManager::getInstance();
    const ResId id(res::ship_rotated, hull, angle+1);
    Ptr<GfxPixmap> pix = mgr.getPixmap(id);
    if (pix)
        return pix;
    pix = mgr.getPixmap(ResId(res::rship, ::getHull(hull).getInternalPictureNumber()));
    if (pix)
        pix = pix->convertToScreenFormat();

    for (int i = 0; i < num_angles; ++i) {
        if (pix == 0) {
            mgr.addPixmap(ResId(res::ship_rotated, hull, i+1), 0);
        } else {
            double rangle = i * (2 * PI / num_angles);
            const int wi = 32, he = 32;
            int x1 = int (wi * std::cos(rangle));
            int y1 = int (wi * std::sin(rangle));
            int x2 = int (he * std::cos(rangle - PI/2));
            int y2 = int (he * std::sin(rangle - PI/2));
            GfxRect new_size = computeStretchRotateBBox(x1, y1, x2, y2);

            Ptr<GfxPixmap> result = new GfxPixmap(new_size.w, new_size.h, *pix);
            doStretchRotateBlit(*pix,
                                *result,
                                GfxRect(*pix),
                                GfxRect(*result),
                                -new_size.x,
                                -new_size.y,
                                x1, y1, x2, y2);
            result->setColorKey(result->getPixel(GfxPoint(0, 0)));

            mgr.addPixmap(ResId(res::ship_rotated, hull, i+1),
                          result->cloneCropped(result->getColorKey()));
        }
    }
    return mgr.getPixmap(id);
}

void
Flak3DShipSprite::draw(GfxCanvas& can)
{
    if (vis.getBattle().getShipByNumber(ship).isAlive()) {
        if (pix) {
            pix->blit(can, getPos().x, getPos().y);
        } else {
            const bool planet = vis.getBattle().getShipByNumber(ship).isPlanet();
            GfxContext ctx;
            ctx.useCanvas(can);
            ctx.useColorScheme(GfxStandardColorScheme::instance);
            ctx.setColor(!planet ? COLOR_BLACK : COLOR_BLUE);
            drawFilledCircle(ctx, getPos().getCenter(), SHIP_RADIUS);
            ctx.setColor(!planet ? COLOR_WHITE : COLOR_GREEN);
            drawCircle(ctx, getPos().getCenter(), SHIP_RADIUS);
        }
    }
}

void
Flak3DShipSprite::tick()
{
    const FlakShip&     sh = vis.getBattle().getShipByNumber(ship);
    const FlakScreenPos sp = vis.getScreenPosition(sh.getPos());

    if (!sh.isAlive())
        return;

    if (!sh.isPlanet()) {
        /* Figure out movement target. If ship has an enemy, let it
           point there. Otherwise, let it point to the origin.
           Originally, ships moved there when they had no enemy. */
        FlakScreenPos ep;
        if (FlakShip* enemy_ptr = sh.fleet_link->status.enemy_ptr)
            ep = vis.getScreenPosition(enemy_ptr->getPos());
        else
            ep = vis.getScreenPosition(FlakPos(0, 0, sh.getPos().z));

        /* Compute angle. If target and ship are on same screen
           position, pick an arbitrary angle. The previous behaviour
           was to simply do nothing in this case, which causes the
           ship sprite to be replaced by the fallback circle when
           people use rewind. */
        int new_angle;
        if (ep.x != sp.x || ep.y != sp.y)
            new_angle = int(std::atan2(double(ep.y - sp.y), double(ep.x - sp.x)) * num_angles / (2*PI) + num_angles + 0.5) % num_angles;
        else
            new_angle = 0;

        /* Turn the ship. We use a timer to limit the amount of turn
           steps per tick. */
        int angle_diff = (num_angles + new_angle - angle) % num_angles;
        int old_angle = angle;
        if (angle == -1) {
            angle = new_angle;
            timer = 0;
        } else if (angle_diff == 0) {
            timer = 0;
        } else if (timer < min_turn_delay) {
            if (vis.isTimeAdvance())
                timer++;
        } else {
            if (angle_diff < num_angles/2) {
                ++angle;
                if (angle >= num_angles)
                    angle = 0;
            } else {
                --angle;
                if (angle < 0)
                    angle = num_angles-1;
            }
            timer = 0;
        }
        if (angle != old_angle) {
            pix = getShipPicture(sh.data.hull, angle);
            update();
        }
        if (pix) {
            setPos(GfxRect(sp.x - pix->getWidth()/2, sp.y - pix->getHeight()/2,
                           pix->getWidth(), pix->getHeight()),
                   sp.z);
            return;
        }
    }

    setPos(GfxRect(sp.x - SHIP_RADIUS, sp.y - SHIP_RADIUS, 2*SHIP_RADIUS+1, 2*SHIP_RADIUS+1), sp.z);
}

/**************************** Flak3DVisualizer ***************************/

Flak3DVisualizer::Flak3DVisualizer(FlakBattle& battle, GfxRect area)
    : FlakFancyVisualizer(battle, area),
      ship_sprites(battle.getNumShips())
{
    Ptr<GfxPixmap> pix = requestPixmap(ResId("flak.background"));
    if (pix)
        getManager().addSprite(new GfxPixmapSprite3D(0, 0, INT_MAX, layer_Scene, pix));
}

Flak3DVisualizer::~Flak3DVisualizer()
{ }

void
Flak3DVisualizer::createSprites(bool is_startup)
{
    if (is_startup)
        computeViewport();

    // create ships
    for (ship_t i = 0, max = getBattle().getNumShips(); i < max; ++i) {
        ship_sprites[i] = new Flak3DShipSprite(*this, i);
        manager.addSprite(ship_sprites[i]);
    }

    // create objects
    for (int player = 1; player <= FLAK_NUM_OWNERS; ++player)
        for (FlakObject* o = getBattle().getPlayerObjects(player); o != 0; o = o->next)
            createObject(*o, FlakBattle::no_ship);
}

Ptr<GfxSprite3D>
Flak3DVisualizer::getShipSprite(ship_t ship)
{
    return ship_sprites[ship];
}

FlakFancyVisualizer::ship_t
Flak3DVisualizer::getShipAtPosition(GfxPoint pt)
{
    ship_t found = FlakBattle::no_ship;
    int found_z = 0;
    for (ship_t i = 0; i < ship_sprites.size(); ++i) {
        if (getBattle().getShipByNumber(i).isAlive()
            && ship_sprites[i]
            && ship_sprites[i]->getPos().contains(pt))
        {
            int this_z = ship_sprites[i]->getPosZ();
            if (found == FlakBattle::no_ship || this_z < found_z) {
                found = i;
                found_z = this_z;
            }
        }
    }
    return found;
}

void
Flak3DVisualizer::createObject(FlakObject& obj, ship_t source)
{
    if (source != FlakBattle::no_ship)
        markShipChanged(source);
    setObjectSprite(obj, FlakSimpleObjectSprite::create(*this, obj), true);
}

/*************************** FlakIsoVisualizer ***************************/

/*! \class FlakIsoVisualizer
    This displays the fight in a simple isometric 3-D view. */

/** Constructor.
    \param battle the fight to display
    \param area screen area to use */
FlakIsoVisualizer::FlakIsoVisualizer(FlakBattle& battle, GfxRect area)
    : Flak3DVisualizer(battle, area),
      center_x(0), center_y(0), range(0)
{
    computeViewport();
}

FlakIsoVisualizer::~FlakIsoVisualizer()
{ }

void
FlakIsoVisualizer::computeViewport()
{
    center_x = center_y = 0;

    int32 min_range = 19000;
    for (fleet_t i = 0; i < fleet_t(the_battle.getNumFleets()); ++i) {
        FlakFleet& f = the_battle.getFleetByNumber(i);
        min_range = std::max(min_range,
                             int32(std::max(std::abs(f.status.position.x - center_x),
                                            std::abs(f.status.position.y - center_y))));
    }
    range = 4*min_range + 2000;
}

FlakBattle::fleet_t
FlakIsoVisualizer::everyTickAction()
{
    return FlakBattle::no_fleet;
}

void
FlakIsoVisualizer::followFleet(fleet_t /*fleet*/)
{ }

FlakFancyVisualizer::EventResult
FlakIsoVisualizer::handleEvent(const UIEvent& evt)
{
    (void) evt;
    return eNotHandled;
}

FlakScreenPos
FlakIsoVisualizer::getScreenPosition(FlakPos fp)
{
    const GfxRect& a = getManager().getBBox();
    return FlakScreenPos(a.getCenter().x + (a.w * (fp.x - center_x - fp.y + center_y) / range),
                         a.getCenter().y - (a.h * (fp.x - center_x + fp.y - center_y) / (3*range)) - fp.z*25000/range,
                         range + fp.x/2 - fp.y/2);
}

/*********************** FlakPerspectiveVisualizer ***********************/

const double t_dist = 50000;
const double t_z_scale = 2000.0/100;
const double t_scale = 240.0/20000;

FlakPerspectiveVisualizer::FlakPerspectiveVisualizer(FlakBattle& battle, GfxRect area)
    : Flak3DVisualizer(battle, area),
      t_cam_azi(0), t_cam_hei(0),
      vis_max_r(),
      t_follow_fleet(FlakBattle::no_fleet),
      auto_cam(true)
{
    computeViewport();
}

FlakPerspectiveVisualizer::~FlakPerspectiveVisualizer()
{ }

void
FlakPerspectiveVisualizer::computeViewport()
{
    int32_t range = 10000;
    for (fleet_t i = 0; i < fleet_t(the_battle.getNumFleets()); ++i) {
        FlakFleet& f = the_battle.getFleetByNumber(i);
        range = std::max(range,
                         std::max(f.status.position.x,
                                  f.status.position.y));
    }
    vis_max_r = range;
}

FlakBattle::fleet_t
FlakPerspectiveVisualizer::everyTickAction()
{
    FlakBattle::fleet_t result = FlakBattle::no_fleet;
    if (auto_cam) {
        double target_azi =  20 * PI / 180;
        double target_he  = -60 * PI / 180;

        if (t_follow_fleet != FlakBattle::no_fleet && !the_battle.getFleetByNumber(t_follow_fleet).isAlive()) {
            fleet_t b = FlakBattle::no_fleet;
            int pl = the_battle.getFleetByNumber(t_follow_fleet).data.player;
            for (fleet_t a = 0; a < fleet_t(the_battle.getNumFleets()); ++a) {
                FlakFleet& f = the_battle.getFleetByNumber(a);
                if (f.isAlive() &&
                    (b == FlakBattle::no_fleet
                     || (f.data.player == pl
                         && the_battle.getFleetByNumber(b).data.player != pl)))
                {
                    b = a;
                }
            }
            if (b != FlakBattle::no_fleet) {
                t_follow_fleet = b;
                result = b;
            }
        }

        if (t_follow_fleet != FlakBattle::no_fleet && the_battle.getFleetByNumber(t_follow_fleet).isAlive()) {
            FlakFleet& f = the_battle.getFleetByNumber(t_follow_fleet);
            int32_t dx = f.status.position.x;
            int32_t dy = f.status.position.y;
            if (FlakShip* ene = f.status.enemy_ptr) {
                FlakFleet* pfp = ene->fleet_link;
                dx -= pfp->status.position.x;
                dy -= pfp->status.position.y;
            }
            if (dx != 0 || dy != 0) {
                target_azi = std::atan2(double(-dx), double(-dy)) + 40 * PI / 180;
            } else {
                return result;
            }
        }
        target_azi = wrapAngle(target_azi - t_cam_azi);
        target_he  = wrapAngle(target_he  - t_cam_hei);
        int a = int(std::abs(target_azi) * 33.0);
        int b = int(std::abs(target_he) * 33.0);
        if (a > b)  b = a;
        if (b < 10) b = 10;
        t_cam_azi = t_cam_azi + target_azi / b;
        t_cam_hei = t_cam_hei + target_he / b;
    }
    return result;
}

void
FlakPerspectiveVisualizer::followFleet(fleet_t fleet)
{
    t_follow_fleet = fleet;
    auto_cam = true;
}

FlakScreenPos
FlakPerspectiveVisualizer::getScreenPosition(FlakPos fp)
{
    double vp = std::sqrt(double(fp.x)*fp.x + double(fp.y)*fp.y);

    double cosa = std::cos(t_cam_azi);
    double sina = std::sin(t_cam_azi);
    double cosg = std::cos(t_cam_hei);
    double sing = std::sin(t_cam_hei);

    double xp, yp, zp;
    if (vp > 0) {
        xp = fp.x * cosa - fp.y * sina;
        yp = fp.x * sina + fp.y * cosa;
    } else {
        xp = 0;
        yp = 0;
    }
    zp = (yp*cosg + 1.0*fp.z*t_z_scale*sing);
    yp = (yp*sing - 1.0*fp.z*t_z_scale*cosg);

    // xp = xp - t_org_x;
    // yp = yp - t_org_y;
    // zp = zp - t_org_z;

    // Now we have the y axis heading straight away, and the x/z plane in front of us.
    // { t_dist / h' = (t_dist + t_cam_dist - yp) / x }
    const GfxRect& a = getManager().getBBox();
    GfxPoint org = a.getCenter();
    return FlakScreenPos(org.x + t_scale * xp * t_dist / (t_dist + vis_max_r - yp),
                         org.y - t_scale * zp * t_dist / (t_dist + vis_max_r - yp),
                         t_dist - yp);
}

FlakFancyVisualizer::EventResult
FlakPerspectiveVisualizer::handleEvent(const UIEvent& evt)
{
    if (evt.type == ev_Key) {
        switch (evt.key.code) {
         case ss_Shift + SDLK_LEFT:
            auto_cam = false;
            t_cam_azi += 0.03;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_RIGHT:
            auto_cam = false;
            t_cam_azi -= 0.03; 
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_UP:
            auto_cam = false;
            t_cam_hei += 0.03;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_DOWN:
            auto_cam = false;
            t_cam_hei -= 0.03;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case 'c':
            auto_cam = !auto_cam;
            return eDontRedrawMe;
         case 'a':
         case '+':
            auto_cam = false;
            if (vis_max_r > 10000) {
                vis_max_r -= 1000;
            }
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case 'z':
         case 'y':
         case '-':
            auto_cam = false;
            if (vis_max_r < 2000000) {
                vis_max_r += 1000;
            }
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
        }
    }
    return eNotHandled;
}
