/**
  *  \file game/fcode.cc
  *  \brief Friendly Code class
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/host.h"
#include "game/racename.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/random.h"

/*! \class GFCode

    This class defines a special friendly code. In particular, it
    associates the code with a condition and description.

    There also are methods for generating random codes and for
    inquiring/managing the special fcode definitions. */

/** Construct empty GFCode object. */
GFCode::GFCode()
    : flags(0)
{ }

/** Construct from fcodes.cc file line. This one handles a line
    that originally contained `c + "," + desc_line'.
    \param c          friendly code
    \param desc_line  description line */
GFCode::GFCode(string_t c, string_t desc_line)
    : code(c)
{
    initFromString(desc_line);
}

/** Destructor. */
GFCode::~GFCode()
{ }

void
GFCode::initFromString(const string_t& desc_line)
{
    string_t flag_str;
    if (!strSplit(desc_line, flag_str, description, ","))
        throw UserException(format(_("Friendly code `%s' lacking description"), code));
    description = strTrim(description);

    races = GPlayerSet::fromInteger(~0U);
    if (!parseFlags(flag_str, "SPBCARU", &flags, &races, true))
        throw UserException(format(_("Malformed flags for friendly code `%s'"), code));
}

/** Get friendly code.
    \return 3-character string. */
const string_t&
GFCode::getFCode() const
{
    return code;
}

/** Get description.
    \return ready-made string. Variables are already filled in. */
string_t
GFCode::getDescription() const
{
    return player_racenames.expandNames(description);
}

/** Check whether the specified friendly code is numeric. Handles all
    host-specific special rules. */
bool
GFCode::isNumeric(const string_t& fc)
{
    string_t::size_type i = 0, end = fc.size();
    if (host.hasSpacePaddedFCodes()) {
        /* Host allows space padding, so trim whitespace. */
        while (end && fc[end-1] == ' ')
            --end;
        while (i < end && fc[i] == ' ')
            ++i;
        if (i == end)
            return false;
    } else {
        /* Must be three digits */
        if (fc.length() != 3)
            return false;
    }

    if (host.hasNegativeFCodes() && fc[i] == '-') {
        /* Host allows sign, and we got one */
        ++i;
        if (i == end)
            return false;
    }

    /* Remainder must be digits */
    while (i < end) {
        if (fc[i] < '0' || fc[i] > '9')
            return false;
        ++i;
    }
    return true;
}

/** Check for universal minefield friendly code.
    \param fc Friendly code to check
    \param tolerant true to accept either case, false to use host's rules */
bool
GFCode::isUniversalMinefieldFCode(const string_t& fc, bool tolerant)
{
    if (host.isTHost())
        tolerant = true;

    return fc.size() == 3
        && (fc[0] == 'm' || (tolerant && fc[0] == 'M'))
        && (fc[1] == 'f' || (tolerant && fc[1] == 'F'));
}

/** Get a friendly code's numeric value. */
int
GFCode::getNumericValue(const string_t& fc)
{
    if (isNumeric(fc)) {
        int n;
        stringValue(fc, n);
        return n;
    } else {
        return 1000;
    }
}

/** Check whether fc is permitted as a "random" code. Random codes must be
    - not special (ignoring case. HOST considers things like "eE7" special)
    - not numeric
    - random enough

    "Not special" means:
    - not listed in special-fcode list (GFCode::isSpecial)
    - not listed in extra-fcode list
    - is not a universal minefield friendly code
    - does not start with "X"
    - does not contain "#" or "?" (those are special to our simulator)

    "Random enough" means it does not contain any duplicate character.

    \todo this special-cases the "mf" and "X" prefixes. Probably
    a better method for defining these prefixes should be designed. */
bool
GFCode::isAllowedRandom(const string_t& fc)
{
    return fc.length() == 3
        && !isUniversalMinefieldFCode(fc, true)
        && fc[0] != 'X' && fc[0] != 'x'
        && fc[1] != fc[0] && fc[2] != fc[0] && fc[1] != fc[2]
        && fc[0] != '#' && fc[1] != '#' && fc[2] != '#'
        && fc[0] != '?' && fc[1] != '?' && fc[2] != '?'
        && !isSpecial(fc, true)
        && !isExtraFC(fc)
        && !isNumeric(fc);
}

/** Generate a random friendly code. See isAllowedRandom() for
    conditions for random fcodes.

    To guarantee termination, this function bails out if it did not
    find a good enough code after a while. This never happened in real
    life. However, lusers can configure their system to trigger the
    termination guard (by defining every numeric character special
    in xtrafcode.txt), so we prefer degradation over crash. */
string_t
GFCode::generateRandomFCode()
{
    int paranoia_counter = 200;
    string_t rv(3, ' ');
    do {
        rv[0] = 33 + myRandom(90);
        rv[1] = 33 + myRandom(90);
        rv[2] = 33 + myRandom(90);
    } while (!isAllowedRandom(rv) && --paranoia_counter);
    return rv;
}

/** Extra fcodes. \todo Since this is compared against UTF-8 encoded
    fcodes, it implicitly assumes that xtrafcode.txt is stored in
    UTF-8 encoding. That's not particularily nice, but on the other
    hand not much of a problem as fcodes are expected to not contain
    extended characters. */
static Buffer<char> extra_fcodes;

/** Check whether fc is a special friendly code according to xtrafcode.txt. */
bool
GFCode::isExtraFC(const string_t& fc)
{
    using namespace std;
    const char* p = extra_fcodes.begin();
    const char* end = extra_fcodes.end();
    while (p < end) {
        while (p < end && isspace(static_cast<unsigned char>(*p)))
            ++p;

        const char* q = p;
        while (p < end && !isspace(static_cast<unsigned char>(*p)))
            ++p;
        if (p != q && string_t::size_type(p-q) <= fc.length() && fc.compare(0, p-q, q, p-q)==0)
            return true;
    }
    return false;
}

/** Clear extra fcodes list.
    \post isExtraFC(x) == false for all codes */
void
GFCode::clearExtraFC()
{
    extra_fcodes.clear();
}

/** Load extra fcodes list. This will append the specified file to the
    current list. To replace the list, clear it first
    (clearExtraFC()). */
void
GFCode::loadExtraFC(Stream& s)
{
    s.readBuffer(extra_fcodes);
    extra_fcodes.append(' ');
}
