/**
  *  \file game/storage/overview.h
  *  \brief Game Data Storage: Directory Overview
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_STORAGE_OVERVIEW_H
#define PCC_V2_GAME_STORAGE_OVERVIEW_H

#include "game/struct.h"            // NUM_PLAYERS
#include "game/playerset.h"
#include "cpluslib/assert.h"
#include "game/host.h"

class Directory;

/** Game Data Storage: Directory Overview.

    This class scans a directory for usable game files.

    For each player, we can find a result file (have_RST) or an
    unpacked game directory (have_Gamedir), or both belonging to the
    same turn. In addition, there can be a matching TRN (have_TRN).

    In addition, we might find a new RST (have_NewRST) or another RST
    (have_OtherRST).

    If there is a conflict, the conflicting data is marked
    (have_Conflict); essentially, this says one can safely load all
    non-conflicting data. */
class GDirectoryOverview {
 public:
    GDirectoryOverview();

    /** Per-player flags. Note that most combinations are possible; see class description. */
    enum {
        have_RST        = 1,       ///< We have a current RST.
        have_TRN        = 2,       ///< We have a current TRN.
        have_Gamedir    = 4,       ///< We have a current game directory.
        have_NewRST     = 8,       ///< We have a new RST (along with game dir).
        have_Conflict   = 16,
        have_OtherRST   = 32
    };
    void scan(Directory& dir, bool rstOnly = false);
    void clear();

    int getPlayerFlags(int pid) const;
    int getGlobalFlags() const;
    const GHost& getHostVersion(int pid) const;

    GPlayerSet getPlayersWhere(int flags) const;
    int getDefaultPlayer() const;

 private:
    int pp_flags[NUM_PLAYERS];
    GHost host_versions[NUM_PLAYERS];
};

/** Get flags (have_XXX) for one player. */
inline int
GDirectoryOverview::getPlayerFlags(int pid) const
{
    ASSERT(pid > 0 && pid <= NUM_PLAYERS);
    return pp_flags[pid-1];
}

/** Get host version for one player. */
inline const GHost&
GDirectoryOverview::getHostVersion(int pid) const
{
    return host_versions[pid-1];
}

#endif
