/**
  *  \file game/vcrdb.h
  *  \brief VCR Database
  *
  *  As of early 2005, we support not only the HOST / PVCR
  *  distinction, but also the FLAK multi-ship combat engine which has
  *  a completely different VCR format. This module provides the
  *  database base classes, as well as the storage engine for
  *  "classic" (one-on-one) VCRs.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_VCRDB_H
#define PCC_V2_GAME_VCRDB_H

#include <vector>
#include "cpluslib/object.h"
#include "cpluslib/string.h"

class GVcrObject;
class ProgressMonitor;

/** Entry in a VCR database. */
class GVcrEntry : public Object {
 public:
    enum {
        NeedQuickOutcome = 1,              ///< Caller needs the result of getOutcome(x).
        NeedCompleteResult = 2             ///< Caller needs the result of getObject(x, true).
    };
    enum Playability {
        IsPlayable,                        ///< Fight is playable.
        IsNotSupported,                    ///< We cannot play it and know why.
        IsDamaged                          ///< We cannot play it and don't know why. Might be host error.
    };

    /** Get number of objects. */
    virtual int getNumObjects() const = 0;

    /** Get an object participating in the fight.
        If the result is requested but not yet known, compute it.
        If the fight cannot be played, but after=true is requested, treat that as after=false.
        \param slot Slot, [0,getNumObjects())
        \param after false to return beginning of fight, true to return after. */
    virtual GVcrObject& getObject(int slot, bool after) = 0;

    /** Get outcome for an object. Can be one of:
        - -1 = unit got destroyed
        - 0 = unit survived or fight not playable
        - 1..NUM_OWNERS = unit got captured by specified player
        \param slot Slot, [0,getNumObjects()) */
    virtual int getOutcome(int slot) = 0;

    /** Check whether this fight is playable. Should operate quickly. */
    virtual Playability getPlayability() = 0;

    /** Prepare the result. Compute this fight's result. Call pm.tick() occasionally
        to give feedback that we're working. If the result is already computed,
        just return.
        \param resultLevel Requested result level (NeedQuickOutcome, NeedCompleteResult,
                           or combination thereof).
        \param pm Progress monitor */
    virtual void prepareResult(int resultLevel, ProgressMonitor& pm) = 0;

    /** Get name of algorithm used to play this fight. */
    virtual string_t getAlgorithmName(bool translate) const = 0;

    /** Check whether Engine/Shield Bonus is active in this fight. */
    virtual bool isESBActive() = 0;
};

/** Basic VCR database. */
class GVcrDatabase : public Object {
 public:
    GVcrDatabase()
        { }
    ~GVcrDatabase()
        { }

    /** Get number of battles in this database. */
    virtual int getNumBattles() const = 0;

    /** Get a VCR entry.
        \param nr Number, [0,getNumBattles()) */
    virtual GVcrEntry& getBattle(int nr) = 0;
};

#endif
