/**
  *  \file int/export.cc
  *  \brief Interpreter: Export
  *
  *  This module implements the definition of an export job
  *  (IntExportFieldList), and the logic to execute such a job
  *  (IntExporter). Actual output is implemented by IntExporter
  *  descendants.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "int/export.h"
#include "cpluslib/format.h"
#include "game/error.h"
#include "game/obj.h"
#include "int/context.h"
#include "int/interror.h"
#include "int/propenum.h"
#include "io/textfile.h"
#include "util/answer.h"
#include "util/misc.h"
#include "util/nls.h"
#include "int/names.h"

static bool
isIdentifierCharacter(char c)
{
    return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '$' || c == '_' || (c >= '0' && c <= '9') || c == '.';
}

/*************************** IntExportFieldList **************************/

IntExportFieldList::IntExportFieldList()
{ }

IntExportFieldList::~IntExportFieldList()
{ }

/** Add list of fields.
    \param spec Comma-separated list of field specifications
    \throw IntError on error */
void
IntExportFieldList::addList(string_t spec)
{
    string_t::size_type n = 0;
    string_t::size_type i = spec.find(',');
    while (i != spec.npos) {
        add(strTrim(string_t(spec, n, i-n)));
        n = i+1;
        i = spec.find(',', n);
    }
    add(strTrim(string_t(spec, n)));
}

/** Add field.
    \param spec Field definition (field name, optionally with '@' and width)
    \throw IntError on error */
void
IntExportFieldList::add(string_t spec)
{
    string_t::size_type m = spec.find('@');
    int width = 0;
    if (m != spec.npos) {
        if (!stringValue(string_t(spec, m+1), width)) {
            throw IntError("Syntax error");
        }
        spec.erase(m);
    }
    string_t s = strTrim(spec);
    if (s.empty()) {
        throw IntError("Syntax error");
    }
    for (size_t i = 0; i < s.size(); ++i) {
        if (!isIdentifierCharacter(s[i])) {
            throw IntError("Syntax error");
        }
    }
    add(items.size(), s, width);
}

/** Add field.
    \param line Add before this line number
    \param name Name of field
    \param width Width of field (0=use default) */
void
IntExportFieldList::add(index_t line, string_t name, int width)
{
    items.insert(items.begin()+line, Item(strUCase(name), width));
}

/** Swap fields.
    \param a,b Positions of fields to swap */
void
IntExportFieldList::swap(index_t a, index_t b)
{
    iter_swap(items.begin()+a, items.begin()+b);
}

/** Delete a field.
    \param line Line to delete. */
void
IntExportFieldList::remove(index_t line)
{
    items.erase(items.begin()+line);
}

/** Change name of a field.
    \param line Position of field
    \param name New name */
void
IntExportFieldList::setFieldName(index_t line, string_t name)
{
    items[line].name = name;
}

/** Change width of a field.
    \param line Position of field
    \param width New width (0=use default) */
void
IntExportFieldList::setFieldWidth(index_t line, int width)
{
    items[line].width = width;
}

/** Change width of a field, relative.
    Essentially, performs setFieldWidth(getFieldWidth() + delta),
    with some special cases for user-interface handling.
    \param index Position of field [0,size())
    \param delta Change */
void
IntExportFieldList::changeFieldWidth(index_t index, int delta)
{
    if (index < items.size()) {
        /* New field width. Clip into range [-999,+999]. When we pass 0, set it to 0 first. */
        Item& it = items[index];
        int ofw = it.width;
        int nfw = std::max(-999, std::min(+999, ofw + delta));
        if (ofw != 0 && (ofw < 0) != (nfw < 0)) {
            nfw = 0;
        }
        it.width = nfw;
    }
}

/** Toggle field's alignment.
    If the field is left-aligned, makes it right-aligned, and vice-versa.
    (This negates the field width.)
    \param index Position of field [0,size()) */
void
IntExportFieldList::toggleFieldAlignment(index_t index)
{
    if (index < items.size()) {
        Item& it = items[index];
        it.width = -it.width;
    }
}

/** Get name of a field.
    \param line Position of field */
const string_t&
IntExportFieldList::getFieldName(index_t line) const
{
    return items[line].name;
}

/** Get width of a field.
    \param line Position of field */
int
IntExportFieldList::getFieldWidth(index_t line) const
{
    return items[line].width;
}

/** Get number of fields. */
IntExportFieldList::index_t
IntExportFieldList::size() const
{
    return items.size();
}

/** Convert field definitions to string. This string can be fed into
    addList() to restore this field list. */
string_t
IntExportFieldList::toString()
{
    string_t result;
    for (index_t i = 0; i < items.size(); ++i) {
        if (i != 0) {
            result += ',';
        }
        result += items[i].name;
        if (items[i].width) {
            result += format("@%d", items[i].width);
        }
    }
    return result;
}

/****************************** IntExporter ******************************/

IntExporter::IntExporter()
{ }

IntExporter::~IntExporter()
{ }

/** Export a field set.
    \param ctx Context to export. Must have objects.
    \param filter Output filter. Will be called with the object Id as questionId to
                  permit/reject export of an object.
    \param fields List of fields to export.

    \throw IntError on error
    \throw GError on error

    Modelled after export.pas::DoExport. */
void
IntExporter::doExport(IntContext* ctx,
                      AnswerProvider& filter,
                      const IntExportFieldList& fields)
{
    // Populate type hints
    struct TypeHintCollector : public IntPropertyAcceptor {
        const IntExportFieldList& fields;
        std::vector<TIntTypeHint> typeHints;
        std::vector<bool> seenTypeHints;

        TypeHintCollector(const IntExportFieldList& fields)
            : fields(fields),
              typeHints(fields.size(), thNone),
              seenTypeHints(fields.size())
            { }
        void addProperty(const string_t& name, TIntTypeHint th)
            {
                for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
                    if (name == fields.getFieldName(i)) {
                        typeHints[i] = th;
                        seenTypeHints[i] = true;
                    }
                }
            }
    };
    TypeHintCollector thc(fields);
    ctx->enumProperties(thc);

    // Anything missing?
    for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
        if (!thc.seenTypeHints[i]) {
            throw IntError("Unknown field, " + fields.getFieldName(i));
        }
    }

    // Do it!
    startTable(fields, &thc.typeHints.front());
    do {
        // Filter objects as requested
        GObject* obj = ctx->getObject();
        if (obj == 0 || filter.ask(obj->getId(), obj->getName(GObject::PlainName))) {
            // Object should be exported
            startRecord();
            for (IntExportFieldList::index_t i = 0; i < fields.size(); ++i) {
                // Obtain value. If anything throws or the lookup fails,
                // the value is left as null.
                std::auto_ptr<IntValue> value;
                try {
                    int32_t adr = ctx->lookup(fields.getFieldName(i));
                    if (adr >= 0) {
                        value.reset(ctx->get(adr));
                    }
                }
                catch (IntError&)
                { }
                addField(value.get(), fields.getFieldName(i), thc.typeHints[i]);
            }
            endRecord();
        }
    } while (ctx->next());
    endTable();
}

bool
parseOutputFormatType(const string_t& str, IntExportOutputType& result)
{
    if (strCaseCmp(str, "text") == 0) {
        result = tText;
        return true;
    } else if (strCaseCmp(str, "table") == 0) {
        result = tTable;
        return true;
    } else if (strCaseCmp(str, "dbf") == 0) {
        result = tDBF;
        return true;
    } else if (strCaseCmp(str, "csv") == 0) {
        result = tCommaSV;
        return true;
    } else if (strCaseCmp(str, "tsv") == 0) {
        result = tTabSV;
        return true;
    } else if (strCaseCmp(str, "ssv") == 0) {
        result = tSemicolonSV;
        return true;
    } else if (strCaseCmp(str, "json") == 0) {
        result = tJSON;
        return true;
    } else if (strCaseCmp(str, "html") == 0) {
        result = tHTML;
        return true;
    } else {
        return false;
    }
}
