/**
  *  \file int/if/buildpred.cc
  *  \brief Class WBuildCommandParser
  */

#include "int/if/buildpred.h"
#include "cpluslib/format.h"
#include "game/actions/baseparts.h"
#include "game/actions/baseship.h"
#include "game/actions/planetbase.h"
#include "game/actions/planetstructs.h"
#include "game/planettransfer.h"
#include "game/spec.h"
#include "game/tables.h"
#include "int/args.h"
#include "int/exec.h"
#include "int/if/baseif.h"
#include "int/if/ifutil.h"
#include "int/value.h"
#include "util/nls.h"
#include "util/unicodechars.h"

namespace {
    void renderAmount(string_t& info, int n, int todo, int added)
    {
        // We might be half-way in a build action
        info += "\n " UTF_BULLET " ";
        if (todo != n) {
            info += format(_("To build: %d/%d"), todo, n);
        } else {
            info += format(_("To build: %d"), n);
        }

        // - We might exceed supported amount (which means command will never finish)
        if (added != todo) {
            info += "\n " UTF_BULLET " ";
            info += format(_("Only %d more supported!"), added);
        }
    }
}

WBuildCommandParser::WBuildCommandParser(GPlanet& pl)
    : m_planet(pl),
      m_limit(),
      m_result()
{ }

// Set limit for commands that build multiple items (e.g. "BuildFactoriesWait").
void
WBuildCommandParser::setLimit(int n)
{
    m_limit = n;
}

// Check task for applicable limit.
void
WBuildCommandParser::loadLimit(const IntAutoTaskEditor& editor, IntExecutionContext& proc, size_t pc)
{
    int v = 0;
    if (editor.getPC() == pc && editor.isInSubroutineCall()) {
        std::auto_ptr<IntValue> value(proc.getVariable("BUILD.REMAINDER"));
        if (IntIntValue* sv = dynamic_cast<IntIntValue*>(value.get())) {
            v = sv->getValue();
        }
    }
    setLimit(v);
}

// Get result.
std::auto_ptr<WBuildCommandParser::Result>
WBuildCommandParser::getResult()
{
    return m_result;
}

// TaskPredictor:
bool
WBuildCommandParser::predictInstruction(const string_t& name, IntArgBlock& args)
{
    try {
        if (name == "BUILDSHIP" || name == "ENQUEUESHIP") {
            handleBuildShip(args);
        } else if (name == "BUILDBASE" || name == "BUILDBASEWAIT") {
            handleBuildBase(args);
        } else if (name == "BUILDDEFENSE" || name == "BUILDDEFENSEWAIT") {
            handleBuildStructure(args, str_Defense);
        } else if (name == "BUILDFACTORIES" || name == "BUILDFACTORIESWAIT") {
            handleBuildStructure(args, str_Factories);
        } else if (name == "BUILDBASEDEFENSE" || name == "BUILDBASEDEFENSEWAIT") {
            handleBuildStructure(args, str_BaseDefense);
        } else if (name == "BUILDMINES" || name == "BUILDMINESWAIT") {
            handleBuildStructure(args, str_Mines);
        } else if (name == "BUILDENGINES" || name == "BUILDENGINESWAIT") {
            handleBuildParts(args, tl_Engine);
        } else if (name == "BUILDHULLS" || name == "BUILDHULLSWAIT") {
            handleBuildParts(args, tl_Hull);
        } else if (name == "BUILDBEAMS" || name == "BUILDBEAMSWAIT") {
            handleBuildParts(args, tl_Beam);
        } else if (name == "BUILDLAUNCHERS" || name == "BUILDLAUNCHERSWAIT") {
            handleBuildParts(args, tl_Torpedo);
        }
    }
    catch (GError&) {
        // TaskPredictor only guards against IntError; let's also guard against GError.
    }
    return true;
}

void
WBuildCommandParser::handleBuildShip(IntArgBlock& args)
{
    TBuildOrder o;
    if (parseBuildShipCommand(args, o)) {
        // Action
        GStarbaseBuildShipAction a(m_planet);
        GPlanetTransfer tx(m_planet);
        GStarbaseBuildTransaction ta(a, tx);
        a.setUsePartsFromStorage(false);
        a.setBuildOrder(o);

        // Result
        m_result.reset(new Result());
        m_result->info = formatBuildOrder(o);
        m_result->cost = ta.getCost();
        m_result->missingAmount = ta.getMissingAmountAsCost();
        m_result->type = OrderType_Ship;
    }
}

void
WBuildCommandParser::handleBuildBase(IntArgBlock& args)
{
    // Only handle "BuildBase" or "BuildBase 1" commands (not "BuildBase 0")
    args.checkArgumentCount(0, 1);
    IntValue* p = args.getNext();
    if (p == 0 || getBoolValue(p) > 0) {
        // Action
        GPlanetTransfer container(m_planet);
        GPlanetBuildStarbaseAction action(m_planet, container, true);

        // Result
        m_result.reset(new Result());
        m_result->info = _("Starbase");
        m_result->cost = action.getCost();
        m_result->missingAmount = action.getMissingAmountAsCost();
    }
}

void
WBuildCommandParser::handleBuildStructure(IntArgBlock& args, TPlanetaryStructure type)
{
    args.checkArgumentCount(1);
    int32_t n;
    if (!checkIntArg(n, args.getNext(), 0, 10000)) {
        return;
    }

    const int todo = getLimitedAmount(n);;
    if (todo > 0) {
        // Action
        GPlanetTransfer container(m_planet);
        GPlanetBuildStructuresAction action(m_planet, container);

        // Result
        const int added = action.addStructures(type, todo, true);
        m_result.reset(new Result());
        m_result->info = getPlanetaryStructureName(type);
        renderAmount(m_result->info, n, todo, added);

        // Cost. Note that this will be partial cost if we're not supporting enough.
        m_result->cost = action.getCost();
        m_result->missingAmount = action.getMissingAmountAsCost();
    }
}

void
WBuildCommandParser::handleBuildParts(IntArgBlock& args, TTechLevel area)
{
    // Parse args
    // (Ignore arg #3, which is the optional "N" flag)
    args.checkArgumentCount(2, 3);
    int type, amount;
    if (!checkIntArg(type, args.getNext(), 0, 10000)) {
        return;
    }
    if (!checkIntArg(amount, args.getNext(), -10000, +10000)) {
        return;
    }

    // Validate
    const GComponent* comp = 0;
    int slot = type;
    switch (area) {
     case tl_Hull:
        slot = getTruehullSlot(m_planet.getOwner(), type);
        comp = (type > 0 && type <= getNumHulls() ? &getHull(type) : 0);
        break;

     case tl_Engine:
        comp = (type > 0 && type <= NUM_ENGINES ? &getEngine(type) : 0);
        break;

     case tl_Beam:
        comp = (type >= 0 && type <= NUM_BEAMS ? &getBeam(type) : 0);
        break;

     case tl_Torpedo:
        comp = (type >= 0 && type <= NUM_TORPS ? &getTorp(type) : 0);
        break;
    }

    const int todo = getLimitedAmount(amount);
    if (slot != 0 && comp != 0 && todo > 0) {
        // Action
        GPlanetTransfer container(m_planet);
        GStarbaseBuildPartsAction action(m_planet);
        GStarbaseBuildTransaction ta(action, container);

        // Result
        const int added = action.addParts(area, slot, todo, true);
        m_result.reset(new Result());
        m_result->info = comp->getName();
        renderAmount(m_result->info, amount, todo, added);
        m_result->cost = ta.getCost();
        m_result->missingAmount = ta.getMissingAmountAsCost();
    }
}

int
WBuildCommandParser::getLimitedAmount(int requested) const
{
    if (m_limit > 0 && m_limit < requested) {
        return m_limit;
    } else {
        return requested;
    }
}
