/**
  *  \file int/if/buildpred.h
  *  \brief Class WBuildCommandParser
  */
#ifndef PCC_V2_INT_IF_BUILDPRED_H
#define PCC_V2_INT_IF_BUILDPRED_H

#include <memory>
#include "game/cost.h"
#include "game/planet.h"
#include "int/autotask.h"
#include "int/predict.h"

/** Parser for "Build" commands.
    This parses a single independant Build command and generates information about it.
    Information includes a textual description of the order, costs, and missing amounts.
    Supported commands include BuildShip, BuildBase, ship component building (BuildHulls etc.),
    and structure building (BuildFactories etc.).

    To use,
    - create;
    - call setLimit or loadLimit to support partially executed commands;
    - call predictInstruction(), indirectly using TaskPredictor::predictStatement();
    - extract result using getResult(). */
class WBuildCommandParser : public IntAutoTaskPredictor {
 public:
    /** Type of order. */
    enum OrderType {
        OrderType_Other,
        OrderType_Ship
    };

    /** Result. */
    struct Result {
        /** Textual information about things to be built.
            In case of a ship, output of formatBuildOrder(). */
        string_t info;

        /** Remaining cost. */
        GCost cost;

        /** Missing resources required to complete command. */
        GCost missingAmount;

        /** True if this is a ship build order. */
        OrderType type;

        Result()
            : info(), cost(), missingAmount(), type(OrderType_Other)
            { }
    };

    /** Constructor.
        @param pl        Planet */
    explicit WBuildCommandParser(GPlanet& pl);

    /** Set limit for commands that build multiple items (e.g.\ "BuildFactoriesWait").
        If nonzero and in range, assumes that this command will build only so many structures
        instead of the amount requested.

        @param n Limit
        @see loadLimit() */
    void setLimit(int n);

    /** Check task for applicable limit.
        If the process is currently executing the given command,
        retrieve the "Build.Remainder" variable.
        This assumes that the command is implemented using a "Do .. While Build.Remainder" loop.

        @param editor TaskEditor
        @param proc   Process
        @param pc     Program counter */
    void loadLimit(const IntAutoTaskEditor& editor, IntExecutionContext& proc, size_t pc);

    /** Get result.
        @return result, if any. */
    std::auto_ptr<Result> getResult();

    // TaskPredictor:
    virtual bool predictInstruction(const string_t& name, IntArgBlock& args);

 private:
    GPlanet& m_planet;
    int m_limit;
    std::auto_ptr<Result> m_result;

    void handleBuildShip(IntArgBlock& args);
    void handleBuildBase(IntArgBlock& args);
    void handleBuildStructure(IntArgBlock& args, TPlanetaryStructure type);
    void handleBuildParts(IntArgBlock& args, TTechLevel area);

    int getLimitedAmount(int requested) const;
};

#endif
