/**
  *  \file int/if/hullif.cc
  *  \brief Interpreter Game Interface: Hull Properties
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/hullif.h"
#include "game/spec.h"
#include "int/interror.h"
#include "cpluslib/misc.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/if/ifutil.h"
#include "int/if/specif.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "int/contextio.h"
#include "int/propenum.h"

/* @q Hull(id:Int):Obj (Function, Context)
   Access hull properties.
   Use as
   | ForEach Hull Do ...
   or
   | With Hull(n) Do ...

   @diff This function was available for use in %With under the name %Hulls() since PCC 1.0.6.
   Do not use the name %Hulls in new code, it is not supported by PCC2; use %Hull instead.

   @see int:index:group:hullproperty|Hull Properties
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFHullGet(IntArgBlock& args)
{
    int32_t id;

    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, getNumHulls()))
        return 0;

    return new IntHullContext(id);
}

/** Implementation of "ForEach Hull".
    \return Context representing first hull */
IntContext*
IFHullMake()
{
    if (getNumHulls() > 0)
        return new IntHullContext(1);
    else
        return 0;
}

int32_t
IFHullDim(int32_t dim)
{
    return dim == 0 ? 1 : getNumHulls()+1;
}

/***************************** IntHullContext ****************************/

namespace {
    enum { HullProperty, ComponentProperty };
}

static const TIntMapping hull_mapping[] = {
    { "BEAM.MAX",         ihpMaxBeams,         HullProperty,      thInt },
    { "CARGO.MAX",        ihpMaxCargo,         HullProperty,      thInt },
    { "CARGO.MAXFUEL",    ihpMaxFuel,          HullProperty,      thInt },
    { "COST.D",           icpCostD,            ComponentProperty, thInt },
    { "COST.M",           icpCostM,            ComponentProperty, thInt },
    { "COST.MC",          icpCostMC,           ComponentProperty, thInt },
    { "COST.STR",         icpCostStr,          ComponentProperty, thString },
    { "COST.T",           icpCostT,            ComponentProperty, thInt },
    { "CREW.NORMAL",      ihpMaxCrew,          HullProperty,      thInt },
    { "ENGINE.COUNT",     ihpNumEngines,       HullProperty,      thInt },
    { "FIGHTER.BAYS",     ihpNumFighterBays,   HullProperty,      thInt },
    { "ID",               icpId,               ComponentProperty, thInt },
    { "IMAGE",            ihpImage,            HullProperty,      thInt },
    { "IMAGE$",           ihpImage2,           HullProperty,      thInt },
    { "MASS",             icpMass,             ComponentProperty, thInt },
    { "NAME",             icpName,             ComponentProperty, thString },
    { "NAME.SHORT",       icpNameShort,        ComponentProperty, thString },
    { "SPECIAL",          ihpSpecial,          HullProperty,      thString },
    { "TECH",             icpTech,             ComponentProperty, thInt },
    { "TECH.HULL",        icpTech,             ComponentProperty, thInt },
    { "TORP.LMAX",        ihpMaxTorpLaunchers, HullProperty,      thInt },
};

/** Constructor.
    \param hid [1,getNumHulls()] Hull Id */
IntHullContext::IntHullContext(int hid)
    : hid(hid)
{ }

IntHullContext::~IntHullContext()
{ }

int32_t
IntHullContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, hull_mapping, countof(hull_mapping));
}

void
IntHullContext::set(int32_t index, IntValue* value)
{
    switch (hull_mapping[index].domain) {
     case HullProperty:
        setHullProperty(getHull(hid), IntHullProperty(hull_mapping[index].index), value);
        break;
     case ComponentProperty:
        setComponentProperty(getHull(hid), IntComponentProperty(hull_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntHullContext::get(int32_t index)
{
    switch (hull_mapping[index].domain) {
     case HullProperty:
        return getHullProperty(getHull(hid), IntHullProperty(hull_mapping[index].index));
     case ComponentProperty:
        return getComponentProperty(getHull(hid), IntComponentProperty(hull_mapping[index].index));
     default:
        return 0;
    }
}

bool
IntHullContext::next()
{
    if (hid < getNumHulls()) {
        ++hid;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntHullContext::clone()
{
    return new IntHullContext(hid);
}

GObject*
IntHullContext::getObject()
{
    return 0;
}

string_t
IntHullContext::toString(bool /*readable*/)
{
    return format("Hull(%d)", hid);
}

void
IntHullContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Hull;
    sv.value = hid;
}

void
IntHullContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(hull_mapping, countof(hull_mapping));
}

/** Get hull property.
    \param h Hull to query
    \param isp Hull property to query */
IntValue*
getHullProperty(const GHull& h, IntHullProperty isp)
{
    switch (isp) {
     case ihpMaxBeams:
        /* @q Beam.Max:Int (Ship Property, Hull Property)
           Maximum number of beams on this ship. */
        return makeIntValue(h.getMaxBeams());
     case ihpMaxCargo:
        /* @q Cargo.Max:Int (Ship Property, Hull Property)
           Maximum cargo on this ship. */
        return makeIntValue(h.getCargo());
     case ihpMaxFuel:
        /* @q Cargo.MaxFuel:Int (Ship Property, Hull Property)
           Maximum fuel on this ship. */
        return makeIntValue(h.getFuel());
     case ihpMaxCrew:
        /* @q Crew.Normal:Int (Ship Property, Hull Property)
           Maximum crew on this ship. */
        return makeIntValue(h.getCrew());
     case ihpNumEngines:
        /* @q Engine.Count:Int (Ship Property, Hull Property)
           Number of engines. */
        return makeIntValue(h.getNumEngines());
     case ihpSpecial:
        /* @q Special:Str (Hull Property)
           Special function summary.
           This is a string identifying the major special functions of this hull.
           The string will contain each letter if and only if the hull
           has the respective ability assigned for all players.
           - "C" (Cloak, including Advanced and Hardened Cloak)
           - "H" (Hyperdrive)
           - "G" (Gravitonic accelerator)
           - "B" (Bioscan, including Full Bioscan)
           - "A" (Alchemy, including Neutronic/Aries Refinery) */
        {
            string_t result;
            const GExpLevelSet levels = GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS);
            const GPlayerSet  players = GPlayerSet::allUpTo(NUM_PLAYERS) - 0;
            if ((h.getPlayersThatCan(hf_Cloak, true, levels)
                 + h.getPlayersThatCan(hf_AdvancedCloak, true, levels)
                 + h.getPlayersThatCan(hf_HardenedCloak, true, levels)).contains(players))
                result += "C";
            if (h.getPlayersThatCan(hf_Hyperdrive, true, levels).contains(players))
                result += "H";
            if (h.getPlayersThatCan(hf_Gravitonic, true, levels).contains(players))
                result += "G";
            if ((h.getPlayersThatCan(hf_Bioscan, true, levels)
                 + h.getPlayersThatCan(hf_FullBioscan, true, levels)).contains(players))
                result += "B";
            if ((h.getPlayersThatCan(hf_MerlinAlchemy, true, levels)
                 + h.getPlayersThatCan(hf_AriesRefinery, true, levels)
                 + h.getPlayersThatCan(hf_NeutronicRefinery, true, levels)).contains(players))
                result += "A";
            return makeStringValue(result);
        }
     case ihpMaxTorpLaunchers:
        /* @q Torp.LMax:Int (Ship Property, Hull Property)
           Maximum number of torpedo launchers on this ship. */
        return makeIntValue(h.getMaxLaunchers());
     case ihpNumFighterBays:
        /* @q Fighter.Bays:Int (Hull Property)
           Number of fighter bays on this ship. */
        return makeIntValue(h.getNumBays());
     case ihpImage:
        /* @q Image:Int (Hull Property)
           Picture number used to display this ship in PCC.
           @assignable */
        return makeIntValue(h.getInternalPictureNumber());
     case ihpImage2:
        /* @q Image$:Int (Hull Property)
           Picture number used to display this ship in planets.exe. */
        return makeIntValue(h.getExternalPictureNumber());
    }
    return 0;
}

/** Set hull property.
    \param h Hull to modify
    \param isp Property to modify
    \param value Value to assign */
void
setHullProperty(GHull& h, IntHullProperty isp, IntValue* value)
{
    int32_t i;
    switch (isp) {
     case ihpImage:
        if (checkIntArg(i, value)) {
            h.setInternalPictureNumber(i);
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}
