/**
  *  \file ui/skin.cc
  *  \brief Skinning and Color Scheme Management
  *
  *  Skins are a means for well-behaved widgets to adapt to their
  *  environment. Oversimplified, a skin is a means for a widget to
  *  draw its background, and to get some good-looking colors. Each
  *  widget has an associated skin object, where (usually) many
  *  widgets share the same instance of an object.
  *
  *  This module implements the abstract and default skin classes
  *  (UISkin, UIDefaultSkin), and defines the commonly-used color
  *  schemes (UIColorSet instances).
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/skin.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/******************************* Color Sets ******************************/

/** Colors used to display something on a light-gray background.
    Examples include a light-gray window, or a standard list box
    entry. */
const UIColorSet GRAY_COLOR_SET = {
    COLOR_BLACK,                // static
    COLOR_GREENSCALE+6,         // green
    COLOR_DARKYELLOWSCALE+6,    // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_BLACK,                // contrast
    COLOR_BLUEGRAY,             // input
    COLOR_BLUEBLACK,            // blue
    COLOR_DARK,                 // faded
    COLOR_BLACK,                // heading
    COLOR_BLUEBLACK,            // selection
    COLOR_WHITE,                // invstatic
    COLOR_GRAY,                 // background
    COLOR_SHIELD+4,             // link
    COLOR_SHIELD+11,            // link shade
    COLOR_SHIELD+8,             // link focus
    { }
};

/** Colors used to display something on a dark-gray background.
    Examples include a dark-gray window, or a control screen. */
const UIColorSet DARK_COLOR_SET = {
    COLOR_GRAY,                 // static
    COLOR_GREEN,                // green
    COLOR_YELLOW,               // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_YELLOW,               // input
    COLOR_BLUEGRAY,             // blue      // ???
    COLOR_GRAYSCALE+4,          // faded     // ???
    COLOR_WHITE,                // heading
    COLOR_YELLOW,               // selection
    COLOR_BLACK,                // invstatic
    COLOR_GRAYSCALE+4,          // background
    COLOR_SHIELD+12,            // link
    COLOR_SHIELD+8,             // link shade
    COLOR_SHIELD+6,             // link focus
    { }
};

/** Colors used to display something on a black background. Examples
    include the selection bar on a list box. */
const UIColorSet BLACK_COLOR_SET = {
    COLOR_WHITE,                // static
    COLOR_GREEN,                // green
    COLOR_YELLOW,               // yellow
    COLOR_RED,                  // red       // FIXME: darker?
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_GRAY,                 // input
    COLOR_BLUEGRAY,             // blue
    COLOR_DARK,                 // faded
    COLOR_WHITE,                // heading
    COLOR_YELLOW,               // selection
    COLOR_BLACK,                // invstatic
    COLOR_BLACK,                // background
    COLOR_SHIELD+12,            // link
    COLOR_SHIELD+8,             // link shade
    COLOR_SHIELD+6,             // link focus
    { }
};

/** Colors used to display something bright on a light-gray
    background. This is used for the selection of a listbox when the
    listbox doesn't have focus. */
const UIColorSet BRIGHT_GRAY_COLOR_SET = {
    COLOR_WHITE,                // static
    COLOR_GREEN,                // green
    COLOR_YELLOW,               // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_GRAY,                 // input
    COLOR_BLUEGRAY,             // blue
    COLOR_DARK,                 // faded
    COLOR_WHITE,                // heading
    COLOR_BLUEBLACK,            // selection
    COLOR_BLACK,                // invstatic
    COLOR_GRAY,                 // background
    COLOR_SHIELD+4,             // link
    COLOR_SHIELD+11,            // link shade
    COLOR_SHIELD+8,             // link focus
    { }
};

/******************************** UIColor ********************************/

bool
UIColor::parse(const string_t& name, Color& result)
{
    struct ColorName {
        const char* name;
        Color color;
    };
    static const ColorName styles[] = {
        { "static",            tc_Static },
        { "green",             tc_Green },
        { "yellow",            tc_Yellow },
        { "red",               tc_Red },
        { "white",             tc_White },
        { "contrast-color",    tc_Contrast },
        { "input-color",       tc_Input },
        { "blue",              tc_Blue },
        { "dim",               tc_Faded },
        { "heading-color",     tc_Heading },
        { "selection-color",   tc_Selection },
        { "inverse-color",     tc_InvStatic },
        { "background-color",  tc_Background },
        { "link-color",        tc_Link },
        { "link-shade-color",  tc_LinkShade },
        { "link-focus-color",  tc_LinkFocus },
    };
    for (size_t i = 0; i < sizeof(styles)/sizeof(styles[0]); ++i) {
        if (name == styles[i].name) {
            result = styles[i].color;
            return true;
        }
    }
    return false;
}

/********************************* UISkin ********************************/

/** \class UISkin
    \brief A widget skin

    Well-behaved widgets can use a "skin" to adapt to their
    environment. Skins essentially provide a means to draw the
    background of a widget, and to draw readable text on it. */

/** Get default skin. This one is used as a fallback if no skin has
    been explicitly set. */
UISkin&
UISkin::getDefaultSkin()
{
    static UISkin d(GRAY_COLOR_SET);
    return d;
}

uint32_t
UISkin::getColor(uint32_t color)
{
    return standard_colors[getColors().table[color]];
}

void
UISkin::drawBackground(GfxContext& ctx, GfxRect area)
{
    drawSolidBar(ctx, area, UIColor::tc_Background);
}
